/**
 * Copyright 2010 Andy Turner, The University of Leeds, UK
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package uk.ac.leeds.ccg.andyt.generic.math;

import java.lang.Math;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.math.MathContext;
import java.math.RoundingMode;

/**
 * For BigDecimal arithmetic.
 * 
 * For the time being it is expected that BigDecimal numbers will not have a
 * precision or scale greater than Integer.MAX_VALUE.
 *
 * Some of the functionality provided in this class may exist in third party
 * libraries with appropriate licenses, but at the time of writing, I have not
 * found them.
 *
 * The aim is for accuracy within a fixed number of decimal places. All methods
 * need to be fully tested to ensure compliance...
 */
//public class Generic_BigDecimal extends BigDecimal {
public class Generic_BigDecimal extends Generic_Number {

    /**
     * Factorial of numbers are quite often wanted (especially in Taylor Series
     * calculations), and when they are, this is for them.
     */
    public Generic_BigInteger _Factorial_Generic_BigInteger;
    /**
     * For storing the Euler–Mascheroni constant correct to a fixed decimal
     * place precision
     */
    public BigDecimal _E;
    /**
     * For storing the PI constant correct to a fixed decimal place precision
     */
    public BigDecimal _PI;
    /**
     * RoundingMode used in calculations
     */
    public RoundingMode _RoundingMode;
    /**
     * Two is quite often wanted as a BigDecimal so it is made available as a
     * constant.
     */
    public static final BigDecimal Two = BigDecimal.valueOf(2);
    ///** Creates a new instance of Generic_BigDecimal */
    //public Generic_BigDecimal( double d1 ) {
    //    super( d1 );
    //}

    /** Creates a new instance of Generic_BigDecimal */
    public Generic_BigDecimal() {
        this(1000);
    }

    public Generic_BigDecimal(int n) {
        _RoundingMode = RoundingMode.HALF_UP;
        init_Factorial_Generic_BigInteger(n);
        init_E(n);
    }

    public final void init_Factorial_Generic_BigInteger(int n) {
        _Factorial_Generic_BigInteger = new Generic_BigInteger();
        _Factorial_Generic_BigInteger.factorial(n);
    }

    public final void init_E(int decimalPlaces) {
        _Factorial_Generic_BigInteger.factorial(decimalPlaces);
        _E = getE(decimalPlaces);
    }

    public final void init_PI(int decimalPlaces) {
        _PI = getPI(decimalPlaces);
    }

    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        Generic_BigDecimal a_Generic_BigDecimal = new Generic_BigDecimal();
        a_Generic_BigDecimal.test();
    }

    private void test() {
        //test_ceilingSignificantDigit();
        //test_floorSignificantDigit();
        //test_getPI();
        //test_getE();
        //test_exp();
        //test_divide();

        test_power();

        //test_log();
        //test_root();
        //test_round();
        //test_sqrt0();
    }

    private void test_divide() {
        int decimalPlaces;
        BigDecimal x;
        BigDecimal y;
        BigDecimal divide;

        decimalPlaces = 6;
        x = new BigDecimal("30.121");
        y = new BigDecimal("0.0121");
        divide = divide(
                x,
                y,
                decimalPlaces,
                _RoundingMode);
        System.out.println(
                "divide(" + x + "," + y + "," + decimalPlaces + "," + _RoundingMode + ") "
                + divide.toPlainString());

        decimalPlaces = 4;
        x = new BigDecimal("30.121");
        y = new BigDecimal("30.121");
        divide = divide(
                x,
                y,
                decimalPlaces,
                _RoundingMode);
        System.out.println(
                "divide(" + x + "," + y + "," + decimalPlaces + "," + _RoundingMode + ") "
                + divide.toPlainString());

        decimalPlaces = 4;
        x = new BigDecimal("0.030121");
        y = new BigDecimal("0.0000121");
        divide = divide(
                x,
                y,
                decimalPlaces,
                _RoundingMode);
        System.out.println(
                "divide(" + x + "," + y + "," + decimalPlaces + "," + _RoundingMode + ") "
                + divide.toPlainString());

        decimalPlaces = 6;
        x = new BigDecimal("0.030121");
        y = new BigDecimal("0.0000121");
        divide = divide(
                x,
                y,
                decimalPlaces,
                _RoundingMode);
        System.out.println(
                "divide(" + x + "," + y + "," + decimalPlaces + "," + _RoundingMode + ") "
                + divide.toPlainString());

    }

    private void test_round() {
        int decimalPlaces;
        BigDecimal x;
        BigDecimal round;
        decimalPlaces = 2;
        x = new BigDecimal("1.0001");
        round = round(x, decimalPlaces, _RoundingMode);
        System.out.println(
                "round(" + x + "," + decimalPlaces + "," + _RoundingMode + ") "
                + round.toPlainString());
        x = new BigDecimal("1.005");
        round = round(x, decimalPlaces, _RoundingMode);
        System.out.println(
                "round(" + x + "," + decimalPlaces + "," + _RoundingMode + ") "
                + round.toPlainString());
        x = new BigDecimal("0.0001");
        round = round(x, decimalPlaces, _RoundingMode);
        System.out.println(
                "round(" + x + "," + decimalPlaces + "," + _RoundingMode + ") "
                + round.toPlainString());
        x = new BigDecimal("0.005");
        round = round(x, decimalPlaces, _RoundingMode);
        System.out.println(
                "round(" + x + "," + decimalPlaces + "," + _RoundingMode + ") "
                + round.toPlainString());
    }

    private void test_ceilingSignificantDigit() {
        BigDecimal x;
        BigDecimal ceilingSignificantDigit;

        x = new BigDecimal("0.0001");
        ceilingSignificantDigit = ceilingSignificantDigit(x);
        System.out.println(
                "ceilingSignificantDigit(" + x + ") "
                + ceilingSignificantDigit.toPlainString());

        x = new BigDecimal("0.00012");
        ceilingSignificantDigit = ceilingSignificantDigit(x);
        System.out.println(
                "ceilingSignificantDigit(" + x + ") "
                + ceilingSignificantDigit.toPlainString());

        x = new BigDecimal("0.0009");
        ceilingSignificantDigit = ceilingSignificantDigit(x);
        System.out.println(
                "ceilingSignificantDigit(" + x + ") "
                + ceilingSignificantDigit.toPlainString());

        x = new BigDecimal("1.00099");
        ceilingSignificantDigit = ceilingSignificantDigit(x);
        System.out.println(
                "ceilingSignificantDigit(" + x + ") "
                + ceilingSignificantDigit.toPlainString());

        x = new BigDecimal("10008798.00099");
        ceilingSignificantDigit = ceilingSignificantDigit(x);
        System.out.println(
                "ceilingSignificantDigit(" + x + ") "
                + ceilingSignificantDigit.toPlainString());


        x = new BigDecimal("-1.00099");
        ceilingSignificantDigit = ceilingSignificantDigit(x);
        System.out.println(
                "ceilingSignificantDigit(" + x + ") "
                + ceilingSignificantDigit.toPlainString());

        x = new BigDecimal("-10008798.00099");
        ceilingSignificantDigit = ceilingSignificantDigit(x);
        System.out.println(
                "ceilingSignificantDigit(" + x + ") "
                + ceilingSignificantDigit.toPlainString());

        x = new BigDecimal("-0.00099");
        ceilingSignificantDigit = ceilingSignificantDigit(x);
        System.out.println(
                "ceilingSignificantDigit(" + x + ") "
                + ceilingSignificantDigit.toPlainString());

        x = new BigDecimal("-0.99");
        ceilingSignificantDigit = ceilingSignificantDigit(x);
        System.out.println(
                "ceilingSignificantDigit(" + x + ") "
                + ceilingSignificantDigit.toPlainString());
    }

    private void test_floorSignificantDigit() {
        BigDecimal x;
        BigDecimal floorSignificantDigit;

        x = new BigDecimal("0.0001");
        floorSignificantDigit = floorSignificantDigit(x);
        System.out.println(
                "floorSignificantDigit(" + x + ") "
                + floorSignificantDigit.toPlainString());

        x = new BigDecimal("0.00012");
        floorSignificantDigit = floorSignificantDigit(x);
        System.out.println(
                "floorSignificantDigit(" + x + ") "
                + floorSignificantDigit.toPlainString());

        x = new BigDecimal("0.0009");
        floorSignificantDigit = floorSignificantDigit(x);
        System.out.println(
                "floorSignificantDigit(" + x + ") "
                + floorSignificantDigit.toPlainString());

        x = new BigDecimal("1.00099");
        floorSignificantDigit = floorSignificantDigit(x);
        System.out.println(
                "floorSignificantDigit(" + x + ") "
                + floorSignificantDigit.toPlainString());

        x = new BigDecimal("10008798.00099");
        floorSignificantDigit = floorSignificantDigit(x);
        System.out.println(
                "floorSignificantDigit(" + x + ") "
                + floorSignificantDigit.toPlainString());


        x = new BigDecimal("-1.00099");
        floorSignificantDigit = floorSignificantDigit(x);
        System.out.println(
                "floorSignificantDigit(" + x + ") "
                + floorSignificantDigit.toPlainString());

        x = new BigDecimal("-10008798.00099");
        floorSignificantDigit = floorSignificantDigit(x);
        System.out.println(
                "floorSignificantDigit(" + x + ") "
                + floorSignificantDigit.toPlainString());

        x = new BigDecimal("-0.00099");
        floorSignificantDigit = floorSignificantDigit(x);
        System.out.println(
                "floorSignificantDigit(" + x + ") "
                + floorSignificantDigit.toPlainString());

        x = new BigDecimal("-0.99");
        floorSignificantDigit = floorSignificantDigit(x);
        System.out.println(
                "floorSignificantDigit(" + x + ") "
                + floorSignificantDigit.toPlainString());
    }

    private void test_getPI() {
        int decimalPlaces;
        double PI_double = Math.PI;
        System.out.println("PI_double " + PI_double);
        BigDecimal PI;
        decimalPlaces = 10;
        PI = getPI(decimalPlaces);
        System.out.println("getPI(" + decimalPlaces + ") " + PI.toPlainString());
        decimalPlaces = 100;
        PI = getPI(decimalPlaces);
        System.out.println("getPI(" + decimalPlaces + ") " + PI.toPlainString());
    }

    private void test_getE() {
        int decimalPlaces;
        double e_double = Math.E;
        System.out.println("e_double " + e_double);
        BigDecimal e;
        decimalPlaces = 10;
        e = getE(decimalPlaces);
        System.out.println("getE(" + decimalPlaces + ") " + e.toPlainString());
        decimalPlaces = 100;
        e = getE(decimalPlaces);
        System.out.println("getE(" + decimalPlaces + ") " + e.toPlainString());
    }

    private void test_power() {
        System.out.println("testpower()");
        BigDecimal x;
        BigInteger y;
        BigDecimal z;
        BigDecimal power;
        int decimalPlaces;
        RoundingMode a_RoundingMode = RoundingMode.HALF_UP;
        long time;
        int n;
        int maxDiv;
        int div;

        decimalPlaces = 10;
//
//        x = new BigDecimal("5.1");
//        n = 13;
//        div = 8;
//        time = System.currentTimeMillis();
//        power = power(
//                x,
//                n,
//                div,
//                decimalPlaces,
//                a_RoundingMode);
//        time = timeReport(time);
//        System.out.println("power0(" + x + "," + n + ") " + power.toPlainString());
//
//        x = new BigDecimal("5.1");
//        n = 117;
//        div = 4;
//        time = System.currentTimeMillis();
//        power = power(
//                x,
//                n,
//                div,
//                decimalPlaces,
//                a_RoundingMode);
//        time = timeReport(time);
//        System.out.println("power0(" + x + "," + n + ") " + power.toPlainString());
//
//        // 1.0180832023
//        x = new BigDecimal("5.1");
//        z = new BigDecimal("0.011");
//        time = System.currentTimeMillis();
//        power = powerLessThanOne(
//                x,
//                z,
//                decimalPlaces,
//                a_RoundingMode);
//        time = timeReport(time);
//        System.out.println("power0(" + x + "," + z + ") " + power.toPlainString());
//
       decimalPlaces = 100;
//
//        x = new BigDecimal("5.1");
//        z = new BigDecimal("0.0122");
//        time = System.currentTimeMillis();
//        power = power(
//                x,
//                z,
//                _E,
//                this,
//                decimalPlaces,
//                a_RoundingMode);
//        time = timeReport(time);
//        System.out.println("power(" + x + ", " + z + ", " + decimalPlaces + ") " + power.toPlainString());
//
//        x = new BigDecimal("5.1");
//        z = new BigDecimal("0.000122");
//        time = System.currentTimeMillis();
//        power = power(
//                x,
//                z,
//                _E,
//                this,
//                decimalPlaces,
//                a_RoundingMode);
//        time = timeReport(time);
//        System.out.println("power(" + x + ", " + z + ", " + decimalPlaces + ") " + power.toPlainString());
//
//        x = new BigDecimal("5.1");
//        z = new BigDecimal("-0.000122");
//        time = System.currentTimeMillis();
//        power = power(
//                x,
//                z,
//                _E,
//                this,
//                decimalPlaces,
//                a_RoundingMode);
//        time = timeReport(time);
//        System.out.println("power(" + x + ", " + z + ", " + decimalPlaces + ") " + power.toPlainString());
//
//        x = new BigDecimal("5.1");
//        z = new BigDecimal("-0.000000002");
//        time = System.currentTimeMillis();
//        power = power(
//                x,
//                z,
//                _E,
//                this,
//                decimalPlaces,
//                a_RoundingMode);
//        time = timeReport(time);
//        System.out.println(
//                "power(" + x + ", " + z + ", " + decimalPlaces + ") "
//                + power.toPlainString());
//
//        decimalPlaces = 100;
//
//        time = System.currentTimeMillis();
//        x = new BigDecimal("5.1");
//        z = new BigDecimal("-0.000000000002");
//        power = power(
//                x,
//                z,
//                _E,
//                this,
//                decimalPlaces,
//                a_RoundingMode);
//        time = timeReport(time);
//        System.out.println(
//                "power(" + x + ", " + z + ", " + decimalPlaces + ") "
//                + power.toPlainString());
//
//        time = System.currentTimeMillis();
//        x = new BigDecimal(_E.toString());
//        z = new BigDecimal(_E.toString());
//        power = power(
//                x,
//                z,
//                _E,
//                this,
//                decimalPlaces,
//                a_RoundingMode);
//        time = timeReport(time);
//        System.out.println(
//                "power(" + x + ", " + z + ", " + decimalPlaces + ") "
//                + power.toPlainString());
//
//        time = System.currentTimeMillis();
//        x = new BigDecimal(_E.toString());
//        z = new BigDecimal("-" + _E.toString());
//        power = power(
//                x,
//                z,
//                _E,
//                this,
//                decimalPlaces,
//                a_RoundingMode);
//        time = timeReport(time);
//        System.out.println(
//                "power(" + x + ", " + z + ", " + decimalPlaces + ") "
//                + power.toPlainString());
//
//
//        time = System.currentTimeMillis();
//        x = new BigDecimal("0.9");
//        z = new BigDecimal("200");
//        power = power(
//                x,
//                z,
//                _E,
//                this,
//                decimalPlaces,
//                a_RoundingMode);
//        time = timeReport(time);
//        System.out.println(
//                "power(" + x + ", " + z + ", " + decimalPlaces + ") "
//                + power.toPlainString());
//
//
//        time = System.currentTimeMillis();
//        x = new BigDecimal("1.1");
//        z = new BigDecimal("200");
//        power = power(
//                x,
//                z,
//                _E,
//                this,
//                decimalPlaces,
//                a_RoundingMode);
//        time = timeReport(time);
//        System.out.println(
//                "power(" + x + ", " + z + ", " + decimalPlaces + ") "
//                + power.toPlainString());
//
//        time = System.currentTimeMillis();
//        x = new BigDecimal("1.0000000000000000000000000000000001");
//        z = new BigDecimal("20000");
//        power = power(
//                x,
//                z,
//                _E,
//                this,
//                decimalPlaces,
//                a_RoundingMode);
//        time = timeReport(time);
//        System.out.println(
//                "power(" + x + ", " + z + ", " + decimalPlaces + ") "
//                + power.toPlainString());
//
//
//        time = System.currentTimeMillis();
//        x = new BigDecimal("1.000000000000000001234567");
//        z = new BigDecimal("2000078764654345654");
//        power = power(
//                x,
//                z,
//                _E,
//                this,
//                decimalPlaces,
//                a_RoundingMode);
//        time = timeReport(time);
//        System.out.println(
//                "power(" + x + ", " + z + ", " + decimalPlaces + ") "
//                + power.toPlainString());
//
//        time = System.currentTimeMillis();
//        x = new BigDecimal("0.9");
//        z = new BigDecimal("0.9");
//        power = power(
//                x,
//                z,
//                _E,
//                this,
//                decimalPlaces,
//                a_RoundingMode);
//        time = timeReport(time);
//        System.out.println(
//                "power(" + x + ", " + z + ", " + decimalPlaces + ") "
//                + power.toPlainString());
//
//       time = System.currentTimeMillis();
//        x = new BigDecimal("0.1");
//        z = new BigDecimal("0.9");
//        power = power(
//                x,
//                z,
//                _E,
//                this,
//                decimalPlaces,
//                a_RoundingMode);
//        time = timeReport(time);
//        System.out.println(
//                "power(" + x + ", " + z + ", " + decimalPlaces + ") "
//                + power.toPlainString());
//
//        time = System.currentTimeMillis();
//        x = new BigDecimal("0.9");
//        z = new BigDecimal("0.1");
//        power = power(
//                x,
//                z,
//                _E,
//                this,
//                decimalPlaces,
//                a_RoundingMode);
//        time = timeReport(time);
//        System.out.println(
//                "power(" + x + ", " + z + ", " + decimalPlaces + ") "
//                + power.toPlainString());
//
//        time = System.currentTimeMillis();
//        x = new BigDecimal("0.9");
//        z = new BigDecimal("0.12");
//        power = power(
//                x,
//                z,
//                _E,
//                this,
//                decimalPlaces,
//                a_RoundingMode);
//        time = timeReport(time);
//        System.out.println(
//                "power(" + x + ", " + z + ", " + decimalPlaces + ") "
//                + power.toPlainString());
//
//        time = System.currentTimeMillis();
//        x = new BigDecimal("0.09");
//        z = new BigDecimal("0.12");
//        power = power(
//                x,
//                z,
//                _E,
//                this,
//                decimalPlaces,
//                a_RoundingMode);
//        time = timeReport(time);
//        System.out.println(
//                "power(" + x + ", " + z + ", " + decimalPlaces + ") "
//                + power.toPlainString());
//
//                time = System.currentTimeMillis();
//        x = new BigDecimal("0.99999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999");
//        z = new BigDecimal("0.999999999");
//        power = power(
//                x,
//                z,
//                _E,
//                this,
//                decimalPlaces,
//                a_RoundingMode);
//        time = timeReport(time);
//        System.out.println(
//                "power(" + x + ", " + z + ", " + decimalPlaces + ") "
//                + power.toPlainString());
//
//        time = System.currentTimeMillis();
//        x = new BigDecimal("0.1");
//        z = new BigDecimal("0.999991");
//        power = power(
//                x,
//                z,
//                _E,
//                this,
//                decimalPlaces,
//                a_RoundingMode);
//        time = timeReport(time);
//        System.out.println(
//                "power(" + x + ", " + z + ", " + decimalPlaces + ") "
//                + power.toPlainString());

        time = System.currentTimeMillis();
        x = new BigDecimal("0.1");
        z = new BigDecimal("0.9999999991");
        power = power(
                x,
                z,
                _E,
                this,
                decimalPlaces,
                a_RoundingMode);
        time = timeReport(time);
        System.out.println(
                "power(" + x + ", " + z + ", " + decimalPlaces + ") "
                + power.toPlainString());

        time = System.currentTimeMillis();
        x = new BigDecimal("0.92");
        double d = 1d/244d;
        z = new BigDecimal(d);
        power = power(
                x,
                z,
                _E,
                this,
                decimalPlaces,
                a_RoundingMode);
        time = timeReport(time);
        System.out.println(
                "power(" + x + ", " + z + ", " + decimalPlaces + ") "
                + power.toPlainString());
        System.out.println(
                "1-power = " + BigDecimal.ONE.subtract(power));
    }

    private void test_log() {
        int base_int;
        BigDecimal x;
        base_int = 10;
        BigDecimal log;
        int decimalPlaces = 10;
        RoundingMode a_RoundingMode = RoundingMode.HALF_UP;
        x = new BigDecimal("10");
        log = log(
                base_int,
                x,
                decimalPlaces,
                a_RoundingMode);
        System.out.println(
                "log(" + base_int + "," + x + ") " + log.toPlainString());
        base_int = 10;
        x = new BigDecimal("100");
        log = log(
                base_int,
                x,
                decimalPlaces,
                a_RoundingMode);
        System.out.println(
                "log(" + base_int + "," + x + ") " + log.toPlainString());
//        x = new BigDecimal("100100");
//        log = log(base_int, x);
//        System.out.println(
//                "log(" + base_int + "," + x + ") " + log.toPlainString());
        x = new BigDecimal("100100.1");
        log = log(
                base_int,
                x,
                decimalPlaces,
                a_RoundingMode);
        System.out.println(
                "log(" + base_int + "," + x + ") " + log.toPlainString());
    }

    private void test_exp() {
        BigDecimal x;
        BigDecimal exp;
        int decimalPlaces;

        System.out.println("_E " + _E);

        decimalPlaces = 10;
        x = new BigDecimal("2");
        exp = exp(
                x,
                _E,
                this,
                decimalPlaces,
                _RoundingMode);
        System.out.println("exp(" + x + ") " + exp.toPlainString());

        x = new BigDecimal("0.02");
        exp = exp(
                x,
                _E,
                this,
                decimalPlaces,
                _RoundingMode);
        System.out.println("exp(" + x + ") " + exp.toPlainString());

        x = new BigDecimal("0.000000012");
        exp = exp(
                x,
                _E,
                this,
                decimalPlaces,
                _RoundingMode);
        System.out.println("exp(" + x + ") " + exp.toPlainString());

        x = new BigDecimal("12345678");
        exp = exp(
                x,
                _E,
                this,
                decimalPlaces,
                _RoundingMode);
        System.out.println("exp(" + x + ") " + exp.toPlainString());

        x = new BigDecimal("12345678.123");
        exp = exp(
                x,
                _E,
                this,
                decimalPlaces,
                _RoundingMode);
        System.out.println("exp(" + x + ") " + exp.toPlainString());




    }

    private void test_root() {
        long time;
        BigDecimal n;
        int root;
        int decimalPlaces;
        RoundingMode a_RoundingMode = RoundingMode.HALF_UP;
        BigDecimal rootn;

//        decimalPlaces = 10;
//        n = new BigDecimal("0.25");
//        root = 42;
//        //BigDecimal rootn = getRoot(n, root, decimalPlaces, maxError);
//        time = System.currentTimeMillis();
//        rootn = root(
//                n,
//                root,
//                decimalPlaces,
//                a_RoundingMode);
//        time = timeReport(time);
//        System.out.println(
//                "n " + n.toString() + " root " + root
//                + " decimalPlaces " + decimalPlaces
//                + " rootn " + rootn.toPlainString());
//        // 0.967531778523892
//        if (rootn.compareTo(new BigDecimal(0.967531778523892)) == 0) {
//            System.out.print("toCompare is correct");
//        } else {
//            System.out.print("toCompare is incorrect rootn should be 0.967531778523892");
//        }
//
//        n = new BigDecimal("27");
//        root = 3;
//        time = System.currentTimeMillis();
//        rootn = root(
//                n,
//                root,
//                decimalPlaces,
//                a_RoundingMode);
//        time = timeReport(time);
//        System.out.println(
//                "n " + n.toString() + " root " + root
//                + " decimalPlaces " + decimalPlaces
//                + " rootn " + rootn.toPlainString());
//        time = timeReport(time);
//        // 9
//        if (rootn.compareTo(new BigDecimal(9)) == 0) {
//            System.out.print("toCompare is correct");
//        } else {
//            System.out.print("toCompare is incorrect rootn should be 9");
//        }
//
        decimalPlaces = 100;
//        rootn = root(
//                n,
//                root,
//                decimalPlaces,
//                a_RoundingMode);
//        System.out.println(
//                "n " + n.toString() + " root " + root
//                + " decimalPlaces " + decimalPlaces
//                + " rootn " + rootn.toPlainString());
//        time = timeReport(time);
//
//        root = 100023;
//        n = new BigDecimal("8904831940328.25");
//        rootn = root(
//                n,
//                root,
//                decimalPlaces,
//                a_RoundingMode);
//        System.out.println(
//                "n " + n.toString() + " root " + root
//                + " decimalPlaces " + decimalPlaces
//                + " rootn " + rootn.toPlainString());
//        time = timeReport(time);
//
//        n = new BigDecimal("0.25");
//        rootn = root(
//                n,
//                root,
//                decimalPlaces,
//                a_RoundingMode);
//        System.out.println(
//                "n " + n.toString() + " root " + root
//                + " decimalPlaces " + decimalPlaces
//                + " rootn " + rootn.toPlainString());
//        time = timeReport(time);
//
//        n = new BigDecimal("0.999");
//        rootn = root(
//                n,
//                root,
//                decimalPlaces,
//                a_RoundingMode);
//        System.out.println(
//                "n " + n.toString() + " root " + root
//                + " decimalPlaces " + decimalPlaces
//                + " rootn " + rootn.toPlainString());
//        time = timeReport(time);
//
//        root = 100023;
//        n = new BigDecimal("89048.25");
//        rootn = root(
//                n,
//                root,
//                decimalPlaces,
//                a_RoundingMode);
//        System.out.println(
//                "n " + n.toString() + " root " + root
//                + " decimalPlaces " + decimalPlaces
//                + " rootn " + rootn.toPlainString());
//        time = timeReport(time);

        BigInteger root_BigInteger;
        root_BigInteger = new BigInteger("1000000000");
        n = new BigDecimal("8904831940328.25");
        time = System.currentTimeMillis();
        rootn = root(
                n,
                root_BigInteger,
                decimalPlaces,
                a_RoundingMode);
        time = timeReport(time);
        System.out.println(
                "n " + n.toString() + " root_BigInteger " + root_BigInteger
                + " decimalPlaces " + decimalPlaces
                + " rootn " + rootn.toPlainString());


        n = new BigDecimal("0.25");
        rootn = root(
                n,
                root_BigInteger,
                decimalPlaces,
                a_RoundingMode);
        time = timeReport(time);
        System.out.println(
                "n " + n.toString() + " root_BigInteger " + root_BigInteger
                + " decimalPlaces " + decimalPlaces
                + " rootn " + rootn.toPlainString());

        n = new BigDecimal("0.999");
        rootn = root(
                n,
                root_BigInteger,
                decimalPlaces,
                a_RoundingMode);
        time = timeReport(time);
        System.out.println(
                "n " + n.toString() + " root_BigInteger " + root_BigInteger
                + " decimalPlaces " + decimalPlaces
                + " rootn " + rootn.toPlainString());

        root_BigInteger = new BigInteger("10000000000");
        n = new BigDecimal("8904831940.25");
        time = System.currentTimeMillis();
        rootn = root(
                n,
                root_BigInteger,
                decimalPlaces,
                a_RoundingMode);
        time = timeReport(time);
        System.out.println(
                "n " + n.toString() + " root_BigInteger " + root_BigInteger
                + " decimalPlaces " + decimalPlaces
                + " rootn " + rootn.toPlainString());


    }

    private void test_sqrt0() {
        // TODO:
        // code application logic here
        //BigDecimal bd0 = new BigDecimal( BigInteger.ONE, 0 );
        //BigDecimal bd0 = new BigDecimal( BigInteger.ONE, 1 );
        //BigDecimal bd0 = new BigDecimal( BigInteger.ONE, 2 );
        //BigDecimal bd0 = new BigDecimal( BigInteger.ONE, 3 );
        //BigDecimal bd0 = new BigDecimal( BigInteger.ONE, 4 );
        //BigDecimal bd0 = new BigDecimal( BigInteger.ONE, 5 );
        //System.out.println( "" + bd0 );
        //BigDecimal bd1 = new BigDecimal( "128902" ).add( bd0 );
        //BigDecimal bd1 = new BigDecimal( "2" );
        //BigDecimal bd1 = new BigDecimal( "0.2" );
        //BigDecimal bd1 = new BigDecimal( "0.0002" );
        long time = System.currentTimeMillis();
        BigDecimal bd1 = new BigDecimal("0.0000000000000000000000000000002");
        int decimalPlaces = 100;
        BigDecimal sqrt0 = sqrt0(bd1, decimalPlaces);
        time = timeReport(time);
        System.out.println(
                "sqrt0(" + bd1 + ", " + decimalPlaces + ") " + sqrt0);
        //sqrt1( bd1, decimalPlaces );
    }

    /**
     * @param x
     * @param y
     * @return x^y
     */
//    @Deprecated
//    public BigDecimal power0(
//            BigDecimal x,
//            BigDecimal y) {
//        BigDecimal toCompare = null;
//        int yscale = y.scale();
//        int tenToPoweryscale_int = (int) Math.pow(10, yscale);
//        BigInteger yunscaled = y.unscaledValue();
//        //System.out.println("yunscaled " + yunscaled);
//        BigDecimal powerxyunscaled = power0(x, yunscaled);
//        //System.out.println("powerxyunscaled " + powerxyunscaled);
//        int decimalPlaces = 10;
//        toCompare = getRoot(powerxyunscaled, tenToPoweryscale_int, decimalPlaces);
//        return toCompare;
//    }
    /**
     *
     * @param x
     * @param y
     * @param decimalPlaces
     * @param a_RoundingMode
     * @return x multiplied by y then rounded;
     */
    public static BigDecimal multiply(
            BigDecimal x,
            BigDecimal y,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        // Deal with special cases?
        BigDecimal result = x.multiply(y);
        return round(result, decimalPlaces, a_RoundingMode);
    }

    /**
     *
     * @param x
     * @param y
     * @param decimalPlaces
     * @param a_RoundingMode
     * @return x multiplied by y then rounded;
     */
    public static BigDecimal multiply(
            BigDecimal x,
            BigInteger y,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        // Deal with special cases?
        BigDecimal y_BigDecimal = new BigDecimal(y);
        return multiply(
                x,
                y_BigDecimal,
                decimalPlaces,
                a_RoundingMode);
    }

    /**
     *
     * @param x
     * @param y
     * @param decimalPlaces
     * @param a_RoundingMode
     * @return x multiplied by y then rounded;
     */
    public static BigDecimal multiply(
            BigInteger x,
            BigDecimal y,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        // Deal with special cases?
        return multiply(
                y,
                x,
                decimalPlaces,
                a_RoundingMode);
    }

    /**
     * @param x
     * @param y
     * @param decimalPlaces
     * @param a_RoundingMode
     * @return x/y then rounded;
     */
    public static BigDecimal divide(
            BigDecimal x,
            BigDecimal y,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        // Deal with special cases
        if (y.compareTo(BigDecimal.ZERO) == 0) {
            throw new ArithmeticException("Attempted division by zero...");
        }
        if (y.compareTo(BigDecimal.ONE) == 0) {
            return new BigDecimal(x.toString());
        }
        if (x.compareTo(y) == 0) {
            return BigDecimal.ONE;
        }
        return divideNoCaseCheck(
                x,
                y,
                decimalPlaces,
                a_RoundingMode);
    }

    /**
     * @param x
     * @param y
     * @param decimalPlaces
     * @param a_RoundingMode
     * @return x/y then rounded;
     */
    private static BigDecimal divideNoCaseCheck(
            BigDecimal x,
            BigDecimal y,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        int integerPartLength = x.divideToIntegralValue(y).toString().length();
        int precision = integerPartLength + decimalPlaces;
        MathContext a_MathContext = new MathContext(precision, a_RoundingMode);
        BigDecimal xDivideY = x.divide(y, a_MathContext);
        return round(xDivideY, decimalPlaces, a_RoundingMode);
    }

    /**
     * @param x
     * @param y
     * @param decimalPlaces
     * @param a_RoundingMode
     * @return x/y then rounded;
     */
    public static BigDecimal divide(
            BigDecimal x,
            BigInteger y,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        // Deal with special cases
        if (y.compareTo(BigInteger.ZERO) == 0) {
            throw new ArithmeticException("Attempted division by zero...");
        }
        if (y.compareTo(BigInteger.ONE) == 0) {
            return new BigDecimal(x.toString());
        }
        BigDecimal y_BigDecimal = new BigDecimal(y);
        if (x.compareTo(y_BigDecimal) == 0) {
            return BigDecimal.ONE;
        }
        return divideNoCaseCheck(
                x,
                y_BigDecimal,
                decimalPlaces,
                a_RoundingMode);
    }

    public static BigDecimal divide(
            BigInteger x,
            BigDecimal y,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        // Deal with special cases
        if (y.compareTo(BigDecimal.ZERO) == 0) {
            throw new ArithmeticException("Attempted division by zero...");
        }
        if (y.compareTo(BigDecimal.ONE) == 0) {
            return new BigDecimal(x.toString());
        }
        BigDecimal x_BigDecimal = new BigDecimal(x);
        if (x_BigDecimal.compareTo(y) == 0) {
            return BigDecimal.ONE;
        }
        return divideNoCaseCheck(
                x_BigDecimal,
                y,
                decimalPlaces,
                a_RoundingMode);
    }

    /**
     *  x^y = e^(y * ln(x))
     * log(u/v)=log(u)-log(v)
     * log(1) = 0;
     * x(a+b) = x^a*x^b
     * @param x
     * @param y 0 < y < 1
     * @param decimalPlaces
     * @return x^y accurate to decimalPlaces
     */
    private static BigDecimal powerLessThanOne(
            BigDecimal x,
            BigDecimal y,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        BigDecimal result;
        BigDecimal epsilon_BigDecimal = new BigDecimal(BigInteger.ONE, decimalPlaces);
//        if (y.compareTo(epsilon_BigDecimal) == -1) {
//            return new BigDecimal(BigInteger.ZERO, decimalPlaces);
//        }
        decimalPlaces += 2;
        BigDecimal y0 = new BigDecimal(y.toString());
        int maxite = y0.precision();
        BigDecimal element;
        BigInteger elementUnscaled;
        BigDecimal elementOne;
        BigInteger elementOneReciprocal;
        BigDecimal root;
        BigDecimal rootMultiple;
        result = BigDecimal.ONE;
        for (int i = 0; i < maxite; i++) {
            element = floorSignificantDigit(y0);
            //elementUnscaled = new BigDecimal(element.unscaledValue());
            elementUnscaled = element.unscaledValue();
            if (elementUnscaled.compareTo(BigInteger.ZERO) == 1) {
                elementOne = divide(
                        element,
                        elementUnscaled,
                        decimalPlaces,
                        a_RoundingMode);
                elementOneReciprocal = reciprocalWillBeInteger(elementOne);
                root = root(
                        x,
                        elementOneReciprocal,
                        decimalPlaces,
                        a_RoundingMode);
                if (root.compareTo(BigDecimal.ZERO) == 1) {
                    rootMultiple = power(
                            root,
                            elementUnscaled,
                            64,
                            decimalPlaces,
                            a_RoundingMode);
                    result = multiply(
                            result,
                            rootMultiple,
                            decimalPlaces,
                            a_RoundingMode);
                }
            }
            y0 = y0.subtract(element);
        }
        return round(result, decimalPlaces - 2, a_RoundingMode);
        //return round(result, decimalPlaces, a_RoundingMode);
    }

    /**
     * @param compare
     * @param x x > 1
     * @param y0 y0 > 1
     * @param div Optimisation parameter...
     * @param decimalPlaces The number of decimal places rounded to if rounding
     * is necessary.
     * @param a_RoundingMode The RoundingMode used in the case rounding is
     * necessary.
     * @return true iff x^y > compare
     */
    public static boolean powerTestAbove(
            BigDecimal compare,
            BigDecimal x,
            BigInteger y0,
            int div,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        if (div < 3) {
            div = 2;
        } else {
            if (div > 256) {
                div = 256;
            }
        }
        BigInteger div_BigInteger = BigInteger.valueOf(div);
        BigDecimal toCompare = BigDecimal.ONE;
        BigDecimal toCompare0;
        // Deal with special case
        if (y0.compareTo(div_BigInteger) == -1) {
            toCompare0 = power(
                    x,
                    y0,
                    div,
                    decimalPlaces,
                    a_RoundingMode);
            if (toCompare0.compareTo(compare) == 1) {
                return true;
            } else {
                return false;
            }
        }
        BigInteger y = new BigInteger(y0.toString());
        BigInteger[] yDivideAndRemainder;
        while (y.compareTo(BigInteger.ONE) == 1) {
            if (y.compareTo(div_BigInteger) == -1) {
                toCompare0 = power(
                        x,
                        y,
                        div,
                        decimalPlaces,
                        a_RoundingMode);
            } else {
                yDivideAndRemainder = y.divideAndRemainder(div_BigInteger);
                boolean powerTest0 = powerTestAbove(
                        compare,
                        x,
                        yDivideAndRemainder[0],
                        div,
                        decimalPlaces,
                        a_RoundingMode);
                if (powerTest0) {
                    return true;
                } else {
                    toCompare0 = power(
                            x,
                            yDivideAndRemainder[0],
                            div,
                            decimalPlaces,
                            a_RoundingMode);
                }
                toCompare0 = power0(
                        toCompare0,
                        div,
                        decimalPlaces,
                        a_RoundingMode);
                if (toCompare.compareTo(compare) == 1) {
                    return true;
                }
                if (yDivideAndRemainder[1].compareTo(BigInteger.ONE) == 1) {
                    BigDecimal resultRemainder = power(
                            x,
                            yDivideAndRemainder[1],
                            div,
                            decimalPlaces,
                            a_RoundingMode);
                    if (resultRemainder.compareTo(compare) == 1) {
                        return true;
                    }
                    toCompare0 = multiply(
                            toCompare0,
                            resultRemainder,
                            decimalPlaces,
                            a_RoundingMode);
                    if (toCompare.compareTo(compare) == 1) {
                        return true;
                    }
                }
            }
            toCompare = multiply(
                    toCompare,
                    toCompare0,
                    decimalPlaces,
                    a_RoundingMode);
            if (toCompare.compareTo(compare) == 1) {
                return true;
            }
            y = y.divide(div_BigInteger);
        }
        return false;
    }

    /**
     * @param compare
     * @param x x > 0
     * @param y0 y0 > 1
     * @param div Optimisation parameter...
     * @param decimalPlaces The number of decimal places rounded to if rounding
     * is necessary.
     * @param a_RoundingMode The RoundingMode used in the case rounding is
     * necessary.
     * @return true iff x^y > compare
     */
    public static boolean powerTestBelow(
            BigDecimal compare,
            BigDecimal x,
            BigInteger y0,
            int div,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        if (div < 3) {
            div = 2;
        } else {
            if (div > 256) {
                div = 256;
            }
        }
        BigInteger div_BigInteger = BigInteger.valueOf(div);
        BigDecimal toCompare = BigDecimal.ONE;
        BigDecimal toCompare0;
        // Deal with special case
        if (y0.compareTo(div_BigInteger) == -1) {
            toCompare0 = power(
                    x,
                    y0,
                    div,
                    decimalPlaces,
                    a_RoundingMode);
            if (toCompare0.compareTo(compare) == -1) {
                return true;
            } else {
                return false;
            }
        }
        BigInteger y = new BigInteger(y0.toString());
        BigInteger[] yDivideAndRemainder;
        while (y.compareTo(BigInteger.ONE) == 1) {
            if (y.compareTo(div_BigInteger) == -1) {
                toCompare0 = power(
                        x,
                        y,
                        div,
                        decimalPlaces,
                        a_RoundingMode);
            } else {
                yDivideAndRemainder = y.divideAndRemainder(div_BigInteger);
                boolean powerTest0 = powerTestAbove(
                        compare,
                        x,
                        yDivideAndRemainder[0],
                        div,
                        decimalPlaces,
                        a_RoundingMode);
                if (powerTest0) {
                    return true;
                } else {
                    toCompare0 = power(
                            x,
                            yDivideAndRemainder[0],
                            div,
                            decimalPlaces,
                            a_RoundingMode);
                }
                toCompare0 = power0(
                        toCompare0,
                        div,
                        decimalPlaces,
                        a_RoundingMode);
                if (toCompare.compareTo(compare) == -1) {
                    return true;
                }
                if (yDivideAndRemainder[1].compareTo(BigInteger.ONE) == 1) {
                    BigDecimal resultRemainder = power(
                            x,
                            yDivideAndRemainder[1],
                            div,
                            decimalPlaces,
                            a_RoundingMode);
                    if (resultRemainder.compareTo(compare) == -1) {
                        return true;
                    }
                    toCompare0 = multiply(
                            toCompare0,
                            resultRemainder,
                            decimalPlaces,
                            a_RoundingMode);
                    if (toCompare.compareTo(compare) == -1) {
                        return true;
                    }
                }
            }
            toCompare = multiply(
                    toCompare,
                    toCompare0,
                    decimalPlaces,
                    a_RoundingMode);
            if (toCompare.compareTo(compare) == -1) {
                return true;
            }
            y = y.divide(div_BigInteger);
        }
        return false;
    }

    /**
     * @param x x > 1
     * @param y x > 1
     * @param div If < 3 this is set to 2. If > 256 set to 256
     * @param decimalPlaces
     * @param a_RoundingMode
     * @return true iff x^y > compare
     */
    @Deprecated
    public static boolean powerTestAbove(
            BigDecimal compare,
            BigDecimal x,
            int y,
            int div,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        if (div < 3) {
            div = 2;
        } else {
            if (div > 256) {
                div = 256;
            }
        }
        BigDecimal toCompare = BigDecimal.ONE;
        BigDecimal toCompare0;
        // Deal with special case
        if (y < div) {
            toCompare0 = power0(
                    x,
                    y,
                    decimalPlaces,
                    a_RoundingMode);
            if (toCompare0.compareTo(compare) == 1) {
                return true;
            } else {
                return false;
            }
        }
        while (y > 1) {
            if (y < div) {
                boolean powerTest0 = powerTestAbove(
                        compare,
                        x,
                        y,
                        div,
                        decimalPlaces,
                        a_RoundingMode);
                if (powerTest0) {
                    return true;
                } else {
                    toCompare0 = power(
                            x,
                            y,
                            div,
                            decimalPlaces,
                            a_RoundingMode);
                }
            } else {
                int party0 = y / div;
                toCompare0 = power(
                        x,
                        party0,
                        div,
                        decimalPlaces,
                        a_RoundingMode);
                if (toCompare.compareTo(compare) == 1) {
                    return true;
                }
                toCompare0 = power0(
                        toCompare0,
                        div,
                        decimalPlaces,
                        a_RoundingMode);
                if (toCompare.compareTo(compare) == 1) {
                    return true;
                }
                int party1 = y % div;
                if (party1 != 0) {
                    BigDecimal resultRemainder = power(
                            x,
                            party1,
                            div,
                            decimalPlaces,
                            a_RoundingMode);
                    if (toCompare.compareTo(compare) == 1) {
                        return true;
                    }
                    toCompare0 = multiply(
                            toCompare0,
                            resultRemainder,
                            decimalPlaces,
                            a_RoundingMode);
                    if (toCompare.compareTo(compare) == 1) {
                        return true;
                    }
                }
            }
            toCompare = multiply(
                    toCompare,
                    toCompare0,
                    decimalPlaces,
                    a_RoundingMode);
            if (toCompare.compareTo(compare) == 1) {
                return true;
            }
            y /= div;
        }
        return false;
    }

    /**
     * @param x x > 0
     * @param y x > 1
     * @param div If < 3 this is set to 2. If > 256 set to 256
     * @param decimalPlaces
     * @param a_RoundingMode
     * @return true iff x^y > compare
     */
    @Deprecated
    public static boolean powerTestBelow(
            BigDecimal compare,
            BigDecimal x,
            int y,
            int div,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        if (div < 3) {
            div = 2;
        } else {
            if (div > 256) {
                div = 256;
            }
        }
        BigDecimal toCompare = BigDecimal.ONE;
        BigDecimal toCompare0;
        // Deal with special case
        if (y < div) {
            toCompare0 = power0(
                    x,
                    y,
                    decimalPlaces,
                    a_RoundingMode);
            if (toCompare0.compareTo(compare) == -1) {
                return true;
            } else {
                return false;
            }
        }
        while (y > 1) {
            if (y < div) {
                boolean powerTest0 = powerTestAbove(
                        compare,
                        x,
                        y,
                        div,
                        decimalPlaces,
                        a_RoundingMode);
                if (powerTest0) {
                    return true;
                } else {
                    toCompare0 = power(
                            x,
                            y,
                            div,
                            decimalPlaces,
                            a_RoundingMode);
                }
            } else {
                int party0 = y / div;
                toCompare0 = power(
                        x,
                        party0,
                        div,
                        decimalPlaces,
                        a_RoundingMode);
                if (toCompare.compareTo(compare) == -1) {
                    return true;
                }
                toCompare0 = power0(
                        toCompare0,
                        div,
                        decimalPlaces,
                        a_RoundingMode);
                if (toCompare.compareTo(compare) == -1) {
                    return true;
                }
                int party1 = y % div;
                if (party1 != 0) {
                    BigDecimal resultRemainder = power(
                            x,
                            party1,
                            div,
                            decimalPlaces,
                            a_RoundingMode);
                    if (toCompare.compareTo(compare) == -1) {
                        return true;
                    }
                    toCompare0 = multiply(
                            toCompare0,
                            resultRemainder,
                            decimalPlaces,
                            a_RoundingMode);
                    if (toCompare.compareTo(compare) == -1) {
                        return true;
                    }
                }
            }
            toCompare = multiply(
                    toCompare,
                    toCompare0,
                    decimalPlaces,
                    a_RoundingMode);
            if (toCompare.compareTo(compare) == -1) {
                return true;
            }
            y /= div;
        }
        return false;
    }

    /**
     * x^y = e^(y * ln(x))
     * @param x
     * @param y
     * @param decimalPlaces The number of decimal places the toCompare must be
     * accurate and rounded to if necessary.
     * @return x^y accurate to decimalPlaces
     */
    public static BigDecimal power(
            BigDecimal x,
            BigDecimal y,
            BigDecimal e,
            Generic_BigDecimal a_Generic_BigDecimal,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        // Deal with special cases
        if (y.compareTo(BigDecimal.ZERO) == 0) {
            return BigDecimal.ONE;
        }
        if (y.compareTo(BigDecimal.ONE) == 0) {
            return x;
        }
        if (x.compareTo(BigDecimal.ONE) == 0) {
            return BigDecimal.ONE;
        }
        BigDecimal result;
        if (y.compareTo(BigDecimal.ZERO) == -1) {
            BigDecimal power = power(
                    x,
                    y.negate(),
                    e,
                    a_Generic_BigDecimal,
                    decimalPlaces + 2,
                    a_RoundingMode);
            result = reciprocal(
                    power,
                    decimalPlaces,
                    a_RoundingMode);
            return result;
        }
        if (y.compareTo(BigDecimal.ONE) == -1) {
            result = powerLessThanOne(
                    x,
                    y,
                    decimalPlaces,
                    a_RoundingMode);
            return result;
        }
        if (y.scale() == 0) {
            return power(
                    x,
                    y.unscaledValue(),
                    256,
                    decimalPlaces,
                    a_RoundingMode);
        }
        // y * ln(x)
        //System.out.println("e " + e);
        BigDecimal lnx;
        if (x.compareTo(BigDecimal.ONE) == 1) {
            lnx = log(
                    e,
                    x,
                    decimalPlaces,
                    a_RoundingMode);
        } else {
            int logPrecision = x.precision() - x.scale() + decimalPlaces;
            lnx = log(
                    e,
                    x,
                    logPrecision,
                    a_RoundingMode);
        }
        //System.out.println("lnx " + lnx.toString());
        BigDecimal ylnx = multiply(
                y,
                lnx,
                decimalPlaces * 2, // This is a safeish guess
                a_RoundingMode);
        //System.out.println("ylnx " + ylnx.toString());
        BigDecimal expylnx = exp(
                ylnx,
                e,
                a_Generic_BigDecimal,
                decimalPlaces,
                a_RoundingMode);
        //System.out.println("expylnx " + expylnx.toString());
        return round(expylnx, decimalPlaces, a_RoundingMode);
    }

    /**
     * 
     * @param x
     * @param y
     * @param decimalPlaces
     * @param a_RoundingMode
     * @return x^y accurate to decimalPlaces
     */
    private static BigDecimal power(
            BigDecimal x,
            long y,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        // The current (Java6) limit for n in x.pow(n)
        // for BigDecimal x and int n is
        // -999999999 < n < 999999999
        if (Math.abs(y) < 999999999) {
            return power(
                    x,
                    (int) y,
                    256,
                    decimalPlaces,
                    a_RoundingMode);
        } else {
            return power(
                    x,
                    BigInteger.valueOf(y),
                    decimalPlaces,
                    256,
                    a_RoundingMode);
        }
    }

    /**
     * @todo set precision
     * @param x
     * @param y
     * @return
     */
    @Deprecated
    private static BigDecimal power0(
            BigDecimal x,
            BigInteger y,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        // Deal with special cases
        if (y.compareTo(BigInteger.ZERO) == 0) {
            return BigDecimal.ONE;
        }
        if (y.compareTo(BigInteger.ONE) == 0) {
            return x;
        }
        if (x.compareTo(BigDecimal.ONE) == 0) {
            return BigDecimal.ONE;
        }
        if (y.compareTo(BigInteger.ZERO) == -1) {
            BigDecimal result;
            BigDecimal power = power4(
                    x,
                    y.negate(),
                    decimalPlaces,
                    a_RoundingMode);
            result = reciprocal(
                    power,
                    decimalPlaces,
                    a_RoundingMode);
            return result;
        }
        return power4(
                x,
                y,
                decimalPlaces,
                a_RoundingMode);
    }

    /**
     * Use {@code power0(BigDecimal,BigInteger,BigInteger,BigInteger,int,RoundingMode)}
     * instead.
     * @param x
     * @param y
     * @param decimalPlaces
     * @param a_RoundingMode
     * @return
     * @deprecated
     */
    @Deprecated
    private static BigDecimal power4(
            BigDecimal x,
            BigInteger y,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        BigInteger Million_BigInteger = BigInteger.valueOf(1000000);
        if (y.compareTo(Million_BigInteger) == -1) {
            return power3(
                    x,
                    y,
                    decimalPlaces,
                    a_RoundingMode);
        }
        BigInteger[] party = y.divideAndRemainder(Generic_BigInteger.Hundred);
        BigDecimal result = power4(
                x,
                party[0],
                decimalPlaces,
                a_RoundingMode);
        result = power3(
                result,
                Million_BigInteger,
                decimalPlaces,
                a_RoundingMode);
        if (party[1].compareTo(BigInteger.ZERO) == 1) {
            BigDecimal resultRemainder = power3(
                    x,
                    party[1],
                    decimalPlaces,
                    a_RoundingMode);
            result = multiply(
                    result,
                    resultRemainder,
                    decimalPlaces,
                    a_RoundingMode);
        }
        return result;
    }

    /**
     * Use {@code power0(BigDecimal,BigInteger,BigInteger,BigInteger,int,RoundingMode)}
     * instead.
     * @param x
     * @param y
     * @param decimalPlaces
     * @param a_RoundingMode
     * @return
     */
    @Deprecated
    private static BigDecimal power3(
            BigDecimal x,
            BigInteger y,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        BigInteger TenThousand_BigInteger = BigInteger.valueOf(10000);
        if (y.compareTo(TenThousand_BigInteger) == -1) {
            return power2(
                    x,
                    y,
                    decimalPlaces,
                    a_RoundingMode);
        }
        BigInteger[] party = y.divideAndRemainder(Generic_BigInteger.Hundred);
        BigDecimal result = power3(
                x,
                party[0],
                decimalPlaces,
                a_RoundingMode);
        result = power2(
                result,
                TenThousand_BigInteger,
                decimalPlaces,
                a_RoundingMode);
        if (party[1].compareTo(BigInteger.ZERO) == 1) {
            BigDecimal resultRemainder = power2(
                    x,
                    party[1],
                    decimalPlaces,
                    a_RoundingMode);
            result = multiply(
                    result,
                    resultRemainder,
                    decimalPlaces,
                    a_RoundingMode);
        }
        return result;
    }

    /**
     * Use {@code power0(BigDecimal,BigInteger,BigInteger,BigInteger,int,RoundingMode)}
     * instead.
     * @param x
     * @param y
     * @param decimalPlaces
     * @param a_RoundingMode
     * @return
     */
    @Deprecated
    private static BigDecimal power2(
            BigDecimal x,
            BigInteger y,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        BigInteger Hundred_BigInteger = BigInteger.valueOf(100);
        if (y.compareTo(Hundred_BigInteger) != 1) {
            return power1(
                    x,
                    y,
                    decimalPlaces,
                    a_RoundingMode);
        }
        BigInteger[] party = y.divideAndRemainder(
                Generic_BigInteger.Hundred);
        BigDecimal result = power2(
                x,
                party[0],
                decimalPlaces,
                a_RoundingMode);
        result = power1(
                result,
                Hundred_BigInteger,
                decimalPlaces,
                a_RoundingMode);
        if (party[1].compareTo(BigInteger.ZERO) == 1) {
            BigDecimal resultRemainder = power1(
                    x,
                    party[1],
                    decimalPlaces,
                    a_RoundingMode);
            result = multiply(
                    result,
                    resultRemainder,
                    decimalPlaces,
                    a_RoundingMode);
        }
        return result;
    }

    /**
     * Use {@code power0(BigDecimal,BigInteger,BigInteger,BigInteger,int,RoundingMode)}
     * instead.
     * @param x
     * @param y
     * @param decimalPlaces
     * @param a_RoundingMode
     * @return
     */
    @Deprecated
    private static BigDecimal power1(
            BigDecimal x,
            BigInteger y,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        BigInteger Four_BigInteger = BigInteger.valueOf(4);
        if (y.compareTo(Four_BigInteger) != 1) {
            return power0(
                    x,
                    y.intValue(),
                    decimalPlaces,
                    a_RoundingMode);
        }
        BigInteger[] party = y.divideAndRemainder(Four_BigInteger);
        BigDecimal result = power1(
                x,
                party[0],
                decimalPlaces,
                a_RoundingMode);
        result = power0(
                result,
                4,
                decimalPlaces,
                a_RoundingMode);
        if (party[1].compareTo(BigInteger.ZERO) == 1) {
            BigDecimal resultRemainder = power0(
                    x,
                    party[1].intValue(),
                    decimalPlaces,
                    a_RoundingMode);
            result = multiply(
                    result,
                    resultRemainder,
                    decimalPlaces,
                    a_RoundingMode);
        }
        return result;
    }

    /**
     * @param x
     * @param y
     * @param div If div < 2 it is set to 2. If div > 256 it is set to 256.
     * @param decimalPlaces
     * @param a_RoundingMode
     * @return x^y accurate to decimalPlaces
     */
    public static BigDecimal power(
            BigDecimal x,
            int y,
            int div,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        if (div < 2) {
            div = 2;
        } else {
            if (div > 256) {
                div = 256;
            }
        }
        // Deal with special cases
        if (y == 0) {
            return BigDecimal.ONE;
        }
        if (y == 1) {
            return new BigDecimal(x.toString());
        }
        if (y < 0) {
            BigDecimal powerminusy = power(
                    x,
                    -y,
                    div,
                    decimalPlaces,
                    a_RoundingMode);
            return reciprocal(
                    powerminusy,
                    decimalPlaces,
                    a_RoundingMode);
        }
        BigDecimal result = BigDecimal.ONE;
        BigDecimal result0;
        if (y < div) {
            result0 = power0(
                    x,
                    y,
                    decimalPlaces,
                    a_RoundingMode);
        } else {
            int party0 = y / div;
            result0 = power(
                    x,
                    party0,
                    div,
                    decimalPlaces,
                    a_RoundingMode);
            result0 = power0(
                    result0,
                    div,
                    decimalPlaces,
                    a_RoundingMode);
            int party1 = y % div;
            if (party1 != 0) {
                BigDecimal resultRemainder = power(
                        x,
                        party1,
                        div / 2,
                        decimalPlaces,
                        a_RoundingMode);
                result0 = multiply(
                        result0,
                        resultRemainder,
                        decimalPlaces,
                        a_RoundingMode);
            }
        }
        result = multiply(
                result,
                result0,
                decimalPlaces,
                a_RoundingMode);
        return result;
    }

    /**
     * @param x
     * @param y0
     * @param div If div < 2 it is set to 2. If div > 256 it is set to 256.
     * @param decimalPlaces
     * @param a_RoundingMode
     * @return x^y0 accurate to decimalPlaces
     */
    public static BigDecimal power(
            BigDecimal x,
            BigInteger y0,
            int div,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        if (div < 2) {
            div = 2;
        } else {
            if (div > 256) {
                div = 256;
            }
        }
        if (y0.compareTo(BigInteger.ZERO) == 0) {
            return BigDecimal.ONE;
        }
        if (y0.compareTo(BigInteger.ONE) == 0) {
            return new BigDecimal(x.toString());
        }
        if (y0.compareTo(BigInteger.ZERO) == -1) {
            BigDecimal powerminusy = power(
                    x,
                    y0.negate(),
                    div,
                    decimalPlaces,
                    a_RoundingMode);
            return reciprocal(
                    powerminusy,
                    decimalPlaces,
                    a_RoundingMode);
        }
        BigInteger div_BigInteger = BigInteger.valueOf(div);
        BigDecimal result = BigDecimal.ONE;
        BigDecimal result0;
        BigInteger y = new BigInteger(y0.toString());
        BigInteger[] yDivideAndRemainder;
        //while (y.compareTo(Generic_BigInteger.Two) == 1) {
        if (y.compareTo(div_BigInteger) == -1) {
            result0 = power0(
                    x,
                    y.intValue(),
                    decimalPlaces,
                    a_RoundingMode);
        } else {
            yDivideAndRemainder = y.divideAndRemainder(div_BigInteger);
            result0 = power(
                    x,
                    yDivideAndRemainder[0],
                    div,
                    decimalPlaces,
                    a_RoundingMode);
            result0 = power0(
                    result0,
                    div,
                    decimalPlaces,
                    a_RoundingMode);
            if (yDivideAndRemainder[1].compareTo(BigInteger.ONE) == 1) {
                BigDecimal resultRemainder = power(
                        x,
                        yDivideAndRemainder[1],
                        div / 2,
                        decimalPlaces,
                        a_RoundingMode);
                result0 = multiply(
                        result0,
                        resultRemainder,
                        decimalPlaces,
                        a_RoundingMode);
            }
        }
        result = multiply(
                result,
                result0,
                decimalPlaces,
                a_RoundingMode);
        return result;
    }

    /**
     * Use {@code power0(BigDecimal,int,int,int,int,RoundingMode)}
     * instead.
     * @param x
     * @param y > 1
     * @param decimalPlaces
     * @param a_RoundingMode
     * @return
     */
    @Deprecated
    private static BigDecimal power1(
            BigDecimal x,
            int y,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        if (y < 4) {
            return power0(
                    x,
                    y,
                    decimalPlaces,
                    a_RoundingMode);
        }
        int party0 = y / 4;
        BigDecimal result = power1(
                x,
                party0,
                decimalPlaces,
                a_RoundingMode);
        result = power0(
                result,
                4,
                decimalPlaces,
                a_RoundingMode);
        int party1 = y % 4;
        if (party1 != 0) {
            BigDecimal resultRemainder = power0(
                    x,
                    party1,
                    decimalPlaces,
                    a_RoundingMode);
            result = multiply(
                    result,
                    resultRemainder,
                    decimalPlaces,
                    a_RoundingMode);
        }
        return result;
    }

    /**
     *
     * @param x
     * @param y > 1 && expected that y is reasonably small (e.g. 4, 
     * probably < 100) otherwise , but for sure less 
     * than < 100
     * @param decimalPlaces
     * @param a_RoundingMode
     * @return
     */
    private static BigDecimal power0(
            BigDecimal x,
            int y,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        // x.pow(y) was returning a mathematically incorrect result for 
        // x = 1.0000000000000000000000000001, y = 100
        // I think that such an x must be a "supernormal" number, but I'm not
        // sure...
        //BigDecimal result = x.pow(y);// Don't attempt to support "supernormal" numbers.
        BigDecimal result = new BigDecimal(x.toString());
//        for (int i = 1; i < y; i++) {
//            result = multiply(
//                    result,
//                    x,
//                    decimalPlaces,
//                    a_RoundingMode);
//        }
//        return result;
        if (y < 10) {
            for (int i = 1; i < y; i++) {
                result = multiply(
                        result,
                        x,
                        decimalPlaces,
                        a_RoundingMode);
            }
            return result;
        }
        // Double up
        int i;
        for (i = 2; i < y; i *= 2) {
            result = multiply(
                    result,
                    result,
                    decimalPlaces,
                    a_RoundingMode);
        }
        // Calculate remainder
        i /= 2;
        BigDecimal remainderResult = power0(
                x,
                y - i,
                decimalPlaces,
                a_RoundingMode);
        // Multiply together
        result = multiply(
                result,
                remainderResult,
                decimalPlaces,
                a_RoundingMode);
        return result;
    }

    /**
     * @param x
     * @param decimalPlace
     * @param a_RoundingMode
     * @return 1/x Accurate to decimalPlace number of decimal places. If x = 0
     * then an IllegalArgumentException is thrown
     */
    public static BigDecimal reciprocal(
            BigDecimal x,
            int decimalPlace,
            RoundingMode a_RoundingMode) {
        BigDecimal result;
        if (x.compareTo(BigDecimal.ZERO) == 0) {
            throw new IllegalArgumentException(
                    "x = 0 in reciprocal(BigDecimal,int,RoundingMode)");
        }
        result = BigDecimal.ONE.divide(
                x,
                decimalPlace,
                a_RoundingMode);
        return round(result, decimalPlace, a_RoundingMode);
    }

    /**
     * @param x
     * @return 1/x if toCompare is an integer or throws a ArithmeticException. If
     * x = 0 then an IllegalArgumentException is thrown
     */
    public static BigInteger reciprocalWillBeInteger(
            BigDecimal x) {
        BigInteger result;
        if (x.compareTo(BigDecimal.ZERO) == 0) {
            throw new IllegalArgumentException(
                    "x = 0 in reciprocalWillBeInteger(BigDecimal,int,RoundingMode)");
        }
        result = BigDecimal.ONE.divide(x, RoundingMode.UNNECESSARY).toBigIntegerExact();
        return result;
    }

//    /**
//     * Adapted from http://everything2.com/index.pl?node_id=946812
//     * @param x
//     * @param decimalPlaces
//     * @return
//     */
//    public BigDecimal lg(
//            BigDecimal x,
//            int decimalPlaces) {
//        final int NUM_OF_DIGITS = decimalPlaces + 2;
//        // need to add one to get the right number of decimalPlaces
//        //  and then add one again to get the next number
//        //  so I can round it correctly.
//        MathContext mc = new MathContext(NUM_OF_DIGITS, RoundingMode.HALF_EVEN);
//        //special conditions:
//        // log(-toCompare) -> exception
//        // log(1) == 0 exactly;
//        // log of a number lessthan one = -log(1/toCompare)
//        if (x.signum() <= 0) {
//            throw new ArithmeticException("log of a negative number! (or zero)");
//        } else if (x.compareTo(BigDecimal.ONE) == 0) {
//            return BigDecimal.ZERO;
//        } else if (x.compareTo(BigDecimal.ONE) < 0) {
//            return (lg((BigDecimal.ONE).divide(x, mc), decimalPlaces)).negate();
//        }
//        StringBuffer sb = new StringBuffer();
//        //number of digits on the left of the decimal point
//        int leftDigits = x.precision() - x.scale();
//        //so, the first digits of the log10 are:
//        sb.append(leftDigits - 1).append(".");
//        //this is the algorithm outlined in the webpage
//        int n = 0;
//        while (n < NUM_OF_DIGITS) {
//            x = (x.movePointLeft(leftDigits - 1)).pow(10, mc);
//            leftDigits = x.precision() - x.scale();
//            sb.append(leftDigits - 1);
//            n++;
//        }
//        BigDecimal ans = new BigDecimal(sb.toString());
//        //Round the number to the correct number of decimal places.
//        ans = ans.round(new MathContext(ans.precision() - ans.scale() + decimalPlaces, RoundingMode.HALF_EVEN));
//        return ans;
//    }
    /**
     * Adapted from http://stackoverflow.com/questions/739532/logarithm-of-a-bigdecimal
     * Pseudocode algorithm for doing a logarithm.
     * Assuming we want log_n of toCompare
     *
     * toCompare = 0;
     * base = x_BigDecimal;
     * input = toCompare;
     * while (input > base)
     *  toCompare++;
     *  input /= base;
     * fraction = 1/2;
     * input *= input;
     * while (((toCompare + fraction) > toCompare) && (input > 1))
     *  if (input > base)
     *    input /= base;
     *    toCompare += fraction;
     *  input *= input;
     *  fraction /= 2.0;
     *
     * The big while loop may seem a bit confusing.
     * On each pass, you can either square your input or you can take the square root of your base; either way, you must divide your fraction by 2. I find squaring the input, and leaving the base alone, to be more accurate.
     * If the input goes to 1, we're through. The log of 1, for any base, is 0, which means we don't need to add any more.
     * if (toCompare + fraction) is not greater than toCompare, then we've hit the limits of precision for our numbering system. We can stop.
     * Obviously, if you're working with a system which has arbitrarily many digits of precision, you will want to put something else in there to limit the loop.
     */
    public static BigDecimal log(
            int base_int,
            BigDecimal x,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        BigDecimal result = BigDecimal.ZERO;
        BigDecimal input = new BigDecimal(x.toString());
        int scale = input.precision() + decimalPlaces;
        int maxite = Math.max(base_int, 10000);
        int ite = 0;
        BigDecimal maxError_BigDecimal = new BigDecimal(
                BigInteger.ONE, decimalPlaces + 1);
        //System.out.println("epsilon_BigDecimal " + epsilon_BigDecimal);
        //System.out.println("scale " + scale);
        BigDecimal two_BigDecimal = new BigDecimal("2");
        BigDecimal base_BigDecimal = new BigDecimal(base_int);
        while (input.compareTo(base_BigDecimal) == 1) {
            result = result.add(BigDecimal.ONE);
            input = input.divide(base_BigDecimal, scale, a_RoundingMode);
        }
        BigDecimal fraction = new BigDecimal("0.5");
        input = input.multiply(input);
        BigDecimal resultplusfraction = result.add(fraction);
        int decimalPlacesSafer = Math.max(decimalPlaces * 2, decimalPlaces + 10); // Not sure this is safe enough, maybe log(base_int,maxite,0) would be?
        while (((resultplusfraction).compareTo(result) == 1)
                && (input.compareTo(BigDecimal.ONE) == 1)) {
            if (input.compareTo(base_BigDecimal) == 1) {
                input = input.divide(base_BigDecimal, scale, a_RoundingMode);
                result = result.add(fraction);
            }
            input = multiply(
                    input,
                    input,
                    decimalPlacesSafer,
                    a_RoundingMode);
            fraction = fraction.divide(two_BigDecimal, scale, a_RoundingMode);
            resultplusfraction = result.add(fraction);
            if (fraction.abs().compareTo(maxError_BigDecimal) == -1) {
                break;
            }
            if (maxite == ite) {
                break;
            }
            ite++;
        }
        return round(result, decimalPlaces, a_RoundingMode);
    }

    /**
     * Only works for x > 0.
     * @param base
     * @param x
     * @param decimalPlaces
     * @param a_RoundingMode
     * @return The log of x to the base base to decimalPlace precision
     */
    public static BigDecimal log(
            BigDecimal base,
            BigDecimal x,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        BigDecimal result = BigDecimal.ZERO;
        // Deal with special cases
        if (x.compareTo(BigDecimal.ZERO) != 1) {
            throw new IllegalArgumentException("x <= 0");
        }
        if (x.compareTo(BigDecimal.ONE) == 0) {
            return BigDecimal.ZERO;
        }
        if (x.compareTo(BigDecimal.ONE) == -1) {
            int xprecision = x.precision();
            BigInteger unscaledx = x.unscaledValue();
            int unscaledxprecision = unscaledx.toString().length();
            int index = xprecision - unscaledxprecision;
            BigDecimal unscaledxmajor = new BigDecimal(unscaledx, index);
            //System.out.print(unscaledxmajor);
            BigDecimal logunscaledxmajor = log(
                    base,
                    unscaledxmajor,
                    decimalPlaces + index,
                    a_RoundingMode);
            result = logunscaledxmajor.subtract(BigDecimal.valueOf(index));
            return round(result, decimalPlaces, a_RoundingMode);
        }
        result = BigDecimal.ZERO;
        BigDecimal input = new BigDecimal(x.toString());
        int scale = input.precision() + decimalPlaces;
        BigInteger maxite = Generic_BigInteger.max(
                x.toBigInteger(),
                BigInteger.valueOf(10000));
        BigInteger ite = BigInteger.ZERO;
        BigDecimal maxError_BigDecimal = new BigDecimal(
                BigInteger.ONE,
                decimalPlaces + 1);
        //System.out.println("epsilon_BigDecimal " + epsilon_BigDecimal);
        //System.out.println("scale " + scale);
        while (input.compareTo(base) == 1) {
            result = result.add(BigDecimal.ONE);
            input = input.divide(base, scale, a_RoundingMode);
        }
        BigDecimal fraction = new BigDecimal("0.5");
        input = input.multiply(input);
        BigDecimal resultplusfraction = result.add(fraction);
        int decimalPlacesSafer = Math.max(decimalPlaces * 2, decimalPlaces + 10); // Not sure this is safe enough, maybe log(base_int,maxite,0) would be?
        while (((resultplusfraction).compareTo(result) == 1)
                && (input.compareTo(BigDecimal.ONE) == 1)) {
            if (input.compareTo(base) == 1) {
                input = input.divide(base, scale, a_RoundingMode);
                result = result.add(fraction);
            }
            input = multiply(
                    input,
                    input,
                    decimalPlacesSafer,
                    a_RoundingMode);
            fraction = fraction.divide(Two, scale, a_RoundingMode);
            resultplusfraction = result.add(fraction);
            if (fraction.abs().compareTo(maxError_BigDecimal) == -1) {
                break;
            }
            if (maxite.compareTo(ite) == 0) {
                break;
            }
            ite = ite.add(BigInteger.ONE);
        }
        return round(result, decimalPlaces, a_RoundingMode);
    }

    /**
     * If x is rounded to a BigDecimal equivalent to 0, the trailing Zeros are
     * not stripped!
     * @param x
     * @param decimalPlaces
     * @param a_RoundingMode
     * @return x rounded to decimalPlaces
     */
    public static BigDecimal round(
            BigDecimal x,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        return x.setScale(decimalPlaces, a_RoundingMode).stripTrailingZeros();
//        BigDecimal toCompare = new BigDecimal(x.toString());
//        // Does the following special case test saves more time than it wastes?
//        if(x.scale() < decimalPlaces){
//            return toCompare;
//        }
//        int precision = (decimalPlaces - 1) + (x.precision() - x.scale());
//        precision = Math.max(precision, 0);
//        MathContext a_MathContext = new MathContext(
//                precision,
//                a_RoundingMode);
//        return toCompare.round(a_MathContext).stripTrailingZeros();
    }

    /**
     * TODO Once completed this can be made public
     */
    private BigDecimal getPIMinDecimalPlaces(
            int decimalPlaces) {
        if (_PI != null) {
            if (_PI.scale() > decimalPlaces) {
                return _PI;
            }
        }
        return getPI(decimalPlaces);
    }

    /**
     * TODO Once completed this can be made public
     * Formula for calculating PI = 4 - 4/3 + 4/5 - 4/7 + 4/9 - 4/11....
     * Because of the time it takes to compute PI, it is preferable to load in a
     * decimally precise version of PI...
     * http://en.wikipedia.org/wiki/Numerical_approximations_of_π
     * Newton:
     * Sum for all k > 0 ((2^k)(k!^2))/(2k+1)!
     * David Chudnovsky and Gregory Chudnovsky:
     * Sum for all k > 0 1/(12((-1^k(6k)!(13591409+54514013k))/(3k)!((k!)^3)640320^(3k+3/2)))
     * @param decimalPlaces
     * @return PI correct to decimalPlaces number of decimal places
     */
    private BigDecimal getPI(
            int decimalPlaces) {
        if (_PI != null) {
            if (_PI.scale() > decimalPlaces) {
                return round(_PI, decimalPlaces, _RoundingMode);
            }
        }
        _PI = BigDecimal.ZERO;
        BigDecimal e = getE(decimalPlaces);
        int maxite = 10;

        // David Chudnovsky and Gregory Chudnovsky:
        // Sum for all k > 0 1/(12((-1^k(6k)!(13591409+54514013k))/(3k)!((k!)^3)640320^(3k+3/2)))
        // Sum for all k > 0 1/(12((-1^k(6k)!d)/(3k!)((k!)^3)640320^(3k+3/2)))
        if (_Factorial_Generic_BigInteger == null) {
            init_Factorial_Generic_BigInteger(maxite);
        } else {
            _Factorial_Generic_BigInteger.factorial(maxite);
        }
        BigDecimal Three = new BigDecimal("3");
        BigDecimal Twelve = new BigDecimal("12");
        BigInteger a = BigInteger.valueOf(13591409);
        BigInteger b = BigInteger.valueOf(54514013);
        BigDecimal c = BigDecimal.valueOf(640320);
        BigDecimal cToh;
        BigInteger d;

        BigDecimal onePointFive = new BigDecimal("1.5");
        BigDecimal h;
        BigInteger k_BigInteger;
        BigDecimal k_BigDecimal;
        BigInteger kfactorial;
        BigInteger kfactorialcubed;
        BigInteger sixkfactorial;
        BigInteger sixkfactoriald;
        BigInteger threekfactorial;
        BigDecimal numerator;
        BigDecimal denominator;
        BigDecimal divisionTimeTwelve;
        // Sum for all k > 0 1/(12((-1^k(6k)!d)/(3k)!((k!)^3)640320^h))
        // Sum for all k > 0 1/(12((-1^k(6k)!d)/(3k)!((k!)^3)c^h))
        // Sum for all k > 0 1/(12((-1^k(6k)!d)/denominator))
        boolean negate = false;
        for (int k_int = 0; k_int < maxite; k_int++) {
            k_BigInteger = BigInteger.valueOf(k_int);
            k_BigDecimal = BigDecimal.valueOf(k_int);
            kfactorial = _Factorial_Generic_BigInteger.factorial(k_int);
            sixkfactorial = _Factorial_Generic_BigInteger.factorial(k_int * 6);
            threekfactorial = _Factorial_Generic_BigInteger.factorial(k_int * 3);
            kfactorialcubed = kfactorial.pow(3);
            d = a.add(b.multiply(k_BigInteger));
            h = (Three.multiply(k_BigDecimal)).add(onePointFive);
            cToh = power(
                    c,
                    h,
                    e,
                    this,
                    decimalPlaces,
                    _RoundingMode);
            denominator = new BigDecimal(threekfactorial.multiply(kfactorialcubed)).multiply(cToh);
            numerator = new BigDecimal(sixkfactorial.multiply(d));
            if (negate) {
                numerator = numerator.negate();
                negate = false;
            } else {
                negate = true;
            }
            divisionTimeTwelve = (numerator.multiply(Twelve)).divide(
                    denominator, decimalPlaces, _RoundingMode);
            _PI = _PI.add(divisionTimeTwelve);
            System.out.println("k_int " + k_int + ", _PI " + _PI);
        }
        _PI = new BigDecimal("4");
        int safeDecimalPlaces = decimalPlaces + 3;
        // This can be set higher, upto nearly Long.MAX_VALUE before BigInteger
        // iteration needs to takeover...
        maxite = Integer.MAX_VALUE;
        //maxite = decimalPlaces * 2;
        //maxite = safeDecimalPlaces;
        BigDecimal tollerance = new BigDecimal(
                BigInteger.ONE, safeDecimalPlaces);
        BigDecimal Four = BigDecimal.valueOf(4);
        BigDecimal fraction;
        boolean add = false;
//        for (long denominator_long = 3; denominator_long < maxite; denominator_long+=4) {
//            fraction = Four.divide(
//                    BigDecimal.valueOf(denominator_int),
//                    safeDecimalPlaces,
//                    _RoundingMode);
//            if (fraction.compareTo(tollerance) == -1) {
//                break;
//            }
//            _PI = _PI.subtract(fraction);
//        }
//        for (long denominator_long = 5; denominator_long < maxite; denominator_long+=4) {
//            fraction = Four.divide(
//                    BigDecimal.valueOf(denominator_int),
//                    safeDecimalPlaces,
//                    _RoundingMode);
//            if (fraction.compareTo(tollerance) == -1) {
//                break;
//            }
//                _PI = _PI.add(fraction);
//        }
        for (long denominator_long = 3;
                denominator_long < maxite;
                denominator_long += 2) {
            fraction = Four.divide(
                    BigDecimal.valueOf(denominator_long),
                    safeDecimalPlaces,
                    _RoundingMode);
            if (fraction.compareTo(tollerance) == -1) {
                break;
            }
            if (add) {
                _PI = _PI.add(fraction);
                add = false;
            } else {
                _PI = _PI.subtract(fraction);
                add = true;
            }
            if (denominator_long % 10000000 == 3) {
                System.out.println("denominator_long " + denominator_long + ", _PI " + _PI);
            }
        }
        _PI = round(_PI, decimalPlaces, _RoundingMode);
        return _PI;
    }

    public BigDecimal getEMinDecimalPlaces(
            int decimalPlaces) {
        if (_E != null) {
            if (_E.scale() > decimalPlaces) {
                return _E;
            }
        }
        return getE(decimalPlaces);
    }

    public BigDecimal getE(
            int decimalPlaces) {
        if (_E != null) {
            if (_E.scale() > decimalPlaces) {
                return round(_E, decimalPlaces, _RoundingMode);
            }
        }
        _E = new BigDecimal("2");
        int safeDecimalPlaces = decimalPlaces + 3;
        //int maxite = 10000;
        //int maxite = decimalPlaces * 2;
        int maxite = safeDecimalPlaces;
        if (_Factorial_Generic_BigInteger == null) {
            init_Factorial_Generic_BigInteger(maxite);
        } else {
            _Factorial_Generic_BigInteger.factorial(maxite);
        }
        BigDecimal factorial;
        BigDecimal oneOverFactorial;
        BigDecimal tollerance = new BigDecimal(
                BigInteger.ONE, safeDecimalPlaces);
        // Use Taylor Series
        Integer factorialKey;
        for (int i = 2; i
                < maxite; i++) {
            factorialKey = new Integer(i);
            factorial = new BigDecimal(
                    _Factorial_Generic_BigInteger._Factorial_TreeMap.get(factorialKey));
            oneOverFactorial = BigDecimal.ONE.divide(
                    factorial, safeDecimalPlaces, _RoundingMode);
            _E = _E.add(oneOverFactorial);
            if (oneOverFactorial.compareTo(tollerance) == -1) {
                break;
            }
        }
        _E = round(_E, decimalPlaces, _RoundingMode);
        return _E;
    }

    /**
     * e^x = 1 + x/1! +x^2/2! + x^3/3! +...
     * @param x
     * @param decimalPlaces
     * @return e^x where e is like _E
     */
    public static BigDecimal exp(
            BigDecimal x,
            BigDecimal e,
            Generic_BigDecimal a_Generic_BigDecimal,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        BigDecimal result;
        // Deal with special cases
        if (x.compareTo(BigDecimal.ZERO) == 0) {
            return BigDecimal.ONE;
        }
        if (x.compareTo(BigDecimal.ONE) == 0) {
            //return getE(decimalPlaces);
            return e;
        }
        int safeDecimalPlaces;
        if (x.compareTo(BigDecimal.ZERO) == -1) {
            safeDecimalPlaces = x.scale() + decimalPlaces;
            BigDecimal exp = exp(
                    x.negate(),
                    e,
                    a_Generic_BigDecimal,
                    safeDecimalPlaces,
                    a_RoundingMode);
            result = reciprocal(
                    exp,
                    decimalPlaces,
                    a_RoundingMode);
            return result;
        }
        if (x.compareTo(BigDecimal.ONE) == -1) {
            result = expLessThanOne(
                    x,
                    e,
                    decimalPlaces,
                    a_RoundingMode);
            return result;
        }
        safeDecimalPlaces = decimalPlaces + 3;
        result = BigDecimal.ONE.add(x);
        int maxite = safeDecimalPlaces;
        if (a_Generic_BigDecimal._Factorial_Generic_BigInteger == null) {
            a_Generic_BigDecimal.init_Factorial_Generic_BigInteger(maxite);
        } else {
            a_Generic_BigDecimal._Factorial_Generic_BigInteger.factorial(maxite);
        }
        BigDecimal factorial;
        BigDecimal xpow;
        BigDecimal xpowOverFactorial;
        BigDecimal tollerance = new BigDecimal(
                BigInteger.ONE, safeDecimalPlaces);
        // Use Taylor Series
        Integer factorialKey;
        BigInteger i_BigInteger;
        for (int i = 2; i < maxite; i++) {
            i_BigInteger = BigInteger.valueOf(i);
            factorialKey = new Integer(i);
            factorial = new BigDecimal(
                    a_Generic_BigDecimal._Factorial_Generic_BigInteger._Factorial_TreeMap.get(factorialKey));
            xpow = power(
                    x,
                    i_BigInteger,
                    64,
                    safeDecimalPlaces,
                    a_RoundingMode);
            xpowOverFactorial = divide(
                    xpow,
                    factorial,
                    safeDecimalPlaces,
                    a_RoundingMode);
            result = result.add(xpowOverFactorial);
            if (xpowOverFactorial.compareTo(tollerance) == -1) {
                break;
            }
        }
        return round(result, decimalPlaces, a_RoundingMode);
    }

    /**
     * @param x 0 < x < 1
     * @param decimalPlaces
     * @return e^x accurate to decimalPlaces
     */
    private static BigDecimal expLessThanOne(
            BigDecimal x,
            BigDecimal e,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        BigDecimal result;
        //BigDecimal epsilon_BigDecimal = new BigDecimal(BigInteger.ONE, decimalPlaces);
        BigDecimal y0 = new BigDecimal(e.toString());
        BigDecimal element;
        BigInteger elementUnscaled;
        BigDecimal elementOne;
        BigInteger elementOneReciprocal;
        BigDecimal root;
        BigDecimal rootMultiple;
        int maxite = y0.precision();
        result = BigDecimal.ONE;
        for (int i = 0; i < maxite; i++) {
            element = floorSignificantDigit(y0);
            elementUnscaled = element.unscaledValue();
            if (elementUnscaled.compareTo(BigInteger.ZERO) == 1) {
                elementOne = divide(
                        element,
                        elementUnscaled,
                        decimalPlaces,
                        a_RoundingMode);
                elementOneReciprocal = reciprocalWillBeInteger(elementOne);
                root = root(
                        x,
                        elementOneReciprocal,
                        decimalPlaces,
                        a_RoundingMode);
                if (root.compareTo(BigDecimal.ZERO) == 1) {
                    rootMultiple = power(
                            root,
                            elementUnscaled,
                            64,
                            decimalPlaces,
                            a_RoundingMode);
                    result = multiply(
                            result,
                            rootMultiple,
                            decimalPlaces,
                            a_RoundingMode);
                }
            }
            y0 = y0.subtract(element);
        }
        return round(result, decimalPlaces, a_RoundingMode);
    }

    //    /**
    //     * Compute the natural logarithm of x to a given scale, x > 0.
    //     * http://stackoverflow.com/questions/739532/logarithm-of-a-bigdecimal
    //     */
    //    public BigDecimal ln(
    //            BigDecimal x,
    //            int scale) {
    //        // Check that x > 0.
    //        if (x.signum() <= 0) {
    //            throw new IllegalArgumentException("toCompare <= 0");
    //        }
    //        // The number of digits to the left of the decimal point.
    //        int magnitude = x.toString().length() - x.scale() - 1;
    //        if (magnitude < 3) {
    //            return lnNewton(x, scale);
    //        } // Compute magnitude*ln(x^(1/magnitude)).
    //        else {
    //            // x^(1/magnitude)
    //            BigDecimal root = getRoot(x, magnitude, scale);
    //            // ln(x^(1/magnitude))
    //            BigDecimal lnRoot = lnNewton(root, scale);
    //            //BigDecimal lnRoot = ln(root, scale);
    //            // magnitude*ln(x^(1/magnitude))
    //            return BigDecimal.valueOf(magnitude).multiply(lnRoot).setScale(scale, BigDecimal.ROUND_HALF_EVEN);
    //        }
    //    }
    /**
     * Compute the natural logarithm of toCompare to a given scale, toCompare > 0.
     * Use Newton's algorithm.http://stackoverflow.com/questions/739532/logarithm-of-a-bigdecimal
     */
    private static BigDecimal lnNewton(
            BigDecimal x,
            BigDecimal e,
            Generic_BigDecimal a_Generic_BigDecimal,
            int scale,
            RoundingMode a_RoundingMode) {
        BigDecimal result = new BigDecimal(x.toString());
        int sp1 = scale + 1;
        BigDecimal term;
        // Convergence tolerance = 5*(10^-(scale+1))
        BigDecimal tolerance = BigDecimal.valueOf(5).movePointLeft(sp1);
        // Loop until the approximations converge
        // (two successive approximations are within the tolerance).
        do {
            // e^toCompare
            BigDecimal exp = exp(
                    result,
                    e,
                    a_Generic_BigDecimal,
                    sp1,
                    a_RoundingMode);
            // (e^toCompare - x)/e^toCompare
            term = exp.subtract(x).divide(exp, sp1, a_RoundingMode);
            // toCompare - (e^toCompare - x)/e^toCompare
            result = result.subtract(term);
            //Thread.yield();
        } while (term.compareTo(tolerance) > 0);
        return result.setScale(scale, a_RoundingMode);
    }

    /**
     * @param x
     * @param y
     * @return the larger (closer to positive infinity) of either x or y
     */
    public static BigDecimal max(
            BigDecimal x,
            BigDecimal y) {
        if (x.compareTo(y) == -1) {
            return new BigDecimal(y.toString());
        } else {
            return new BigDecimal(x.toString());
        }
    }

    /**
     * @param x
     * @param y
     * @return the larger (closer to negative infinity) of either x or y
     */
    public static BigDecimal min(
            BigDecimal x,
            BigDecimal y) {
        if (x.compareTo(y) == -1) {
            return new BigDecimal(x.toString());
        } else {
            return new BigDecimal(y.toString());
        }
    }

    public static BigDecimal floorSignificantDigit(
            BigDecimal x) {
        // Deal with special cases
        if (x.compareTo(BigDecimal.ZERO) == -1) {
            BigDecimal result;
            result = ceilingSignificantDigit(x.negate());
            return result.negate();
        }
        if (x.compareTo(BigDecimal.ZERO) == 0) {
            return BigDecimal.ZERO;
        }
        BigDecimal result;
        BigInteger xUnscaledValue = x.unscaledValue();
        String xUnscaledValueString = xUnscaledValue.toString();
        int scale = x.scale() - xUnscaledValueString.length() + 1;
        int biggest = new Integer(xUnscaledValueString.substring(0, 1));
        if (biggest == 0) {
            result = new BigDecimal(BigInteger.ONE, scale);
        } else {
            result = new BigDecimal(BigInteger.valueOf(biggest), scale);
        }
        return result.stripTrailingZeros();
    }

    public static BigDecimal ceilingSignificantDigit(
            BigDecimal x) {
        // Deal with special cases
        if (x.compareTo(BigDecimal.ZERO) == -1) {
            BigDecimal result;
            result = floorSignificantDigit(x.negate());
            return result.negate();
        }
        if (x.compareTo(BigDecimal.ZERO) == 0) {
            return BigDecimal.ZERO;
        }
        BigDecimal result;
        BigInteger xUnscaledValue = x.unscaledValue();
        String xUnscaledValueString = xUnscaledValue.toString();
        int scale = x.scale() - xUnscaledValueString.length() + 1;
        int biggest = new Integer(xUnscaledValueString.substring(0, 1));
        biggest++;
        result = new BigDecimal(BigInteger.valueOf(biggest), scale);
        return result.stripTrailingZeros();
    }

    /**
     * @param x
     * @param root
     * @param decimalPlaces
     * @param a_RoundingMode
     * @return The root of x to decimalPlaces precision
     */
    public static BigDecimal root(
            BigDecimal x,
            BigInteger root,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        // The current (Java6) limit for n in x.pow(n)
        // for BigDecimal x and int n is
        // -999999999 < n < 999999999
        if (root.compareTo(BigInteger.valueOf(999999999)) != 1) {
            return root(
                    x,
                    root.intValue(),
                    decimalPlaces,
                    a_RoundingMode);
        } else {
            // Deal with special cases
            if (x.compareTo(BigDecimal.ZERO) == -1) {
                throw new IllegalArgumentException(
                        "x < 0 in " + Generic_BigDecimal.class
                        + ".root(BigDecimal,int,int,RoundingMode)");
            }
            if (x.compareTo(BigDecimal.ZERO) == 0) {
                return BigDecimal.ZERO;
            }
            if (x.compareTo(BigDecimal.ONE) == 0) {
                return BigDecimal.ONE;
            }
            if (root.compareTo(BigInteger.ZERO) == 0) {
                throw new IllegalArgumentException(
                        "root = 0 in " + Generic_BigDecimal.class
                        + ".root(BigDecimal,BigInteger,int,RoundingMode)");
            }
            if (root.compareTo(BigInteger.ONE) == 0) {
                return new BigDecimal(x.toString());
            }
            if (x.compareTo(BigDecimal.ONE) == -1) {
                return rootLessThanOne(
                        x,
                        root,
                        decimalPlaces,
                        a_RoundingMode);
            }
            //int safeDecimalPlaces = decimalPlaces + 5;
            BigDecimal epsilon_BigDecimal =
                    new BigDecimal(BigInteger.ONE, decimalPlaces);
            // Check there is a root in the precision and return 0 if not
            BigDecimal comparator = BigDecimal.ONE.add(epsilon_BigDecimal);
            boolean powerTest = powerTestAbove(
                    x,
                    comparator,
                    root,
                    256,
                    decimalPlaces,
                    a_RoundingMode);
            if (powerTest) {
                System.out.println("No root in the precision... ");
                return BigDecimal.ZERO;
            }
            BigDecimal result = new BigDecimal("1");
            int rootInitialisationMaxIte = 10;
            result = rootInitialisation(
                    x,
                    root,
                    epsilon_BigDecimal,
                    rootInitialisationMaxIte,
                    decimalPlaces,
                    a_RoundingMode);
            // Newton Raphson
            int newtonRaphsonMaxite = 100;
            return newtonRaphson(
                    x,
                    result,
                    root,
                    epsilon_BigDecimal,
                    newtonRaphsonMaxite,
                    decimalPlaces,
                    a_RoundingMode);
        }
    }

    /**
     * Uses MathContext division
     * @param x
     * @param result0
     * @param root
     * @param epsilon
     * @param maxite
     * @param decimalPlaces
     * @param a_RoundingMode
     * @return Re-approximation of result0 using Newton Raphson method
     */
    public static BigDecimal newtonRaphson(
            BigDecimal x,
            BigDecimal result0,
            BigInteger root,
            BigDecimal epsilon,
            int maxite,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        BigDecimal result = new BigDecimal(result0.toString());
        int precision = decimalPlaces + 5;
        MathContext a_MathContext = new MathContext(precision);
        BigDecimal previousResult_BigDecimal;
        BigDecimal previousResult_BigDecimal0;
        BigDecimal resultipowroot;
        BigDecimal resultipowrootsubtract1;
        BigDecimal divisor;
        BigInteger rootsubtract1 = root.subtract(BigInteger.ONE);

        boolean maxiteReached = true;
        //resultipowroot = toCompare.pow(root, a_MathContext);
        //resultipowrootsubtract1 = toCompare.pow(root - 1, a_MathContext);
        //divisor = root_BigDecimal.multiply(resultipowrootsubtract1, a_MathContext);

        // Initialise toCompare and previousResult_BigDecimal
        previousResult_BigDecimal = new BigDecimal("1");
        previousResult_BigDecimal0 = new BigDecimal("1");

        // Newton Raphson
        for (int i = 0; i < maxite; i++) {
            resultipowroot = power(
                    result,
                    root,
                    64,
                    precision,
                    a_RoundingMode);
            resultipowrootsubtract1 = power(
                    result,
                    rootsubtract1,
                    64,
                    precision,
                    a_RoundingMode);
            divisor = multiply(
                    root,
                    resultipowrootsubtract1,
                    precision,
                    a_RoundingMode);
            result = result.subtract(
                    resultipowroot.subtract(x).divide(divisor, a_MathContext));
            // Test for convergence
            if (previousResult_BigDecimal.subtract(result).abs().compareTo(epsilon) != 1) {
                //System.out.println("Root found in iteration " + i + " out of " + maxite);
                maxiteReached = false;
                break;
            }
            previousResult_BigDecimal0 = new BigDecimal(previousResult_BigDecimal.toString());
            previousResult_BigDecimal = new BigDecimal(result.toString());
        }
        if (maxiteReached) {
            System.out.println(
                    "maxite reached without finding root in "
                    + Generic_BigDecimal.class.getName() + ".newtonRaphson(...)" 
                    + " previousResult_BigDecimal0.subtract(result).abs() "
                    + previousResult_BigDecimal0.subtract(result).abs());
        }
        return result;
    }

    /**
     * Uses decimalPlace division
     * @param x
     * @param result0
     * @param root_BigInteger
     * @param epsilon_BigDecimal
     * @param maxite
     * @param decimalPlaces
     * @param a_RoundingMode
     * @return Re-approximation of result0 using Newton Raphson method
     */
    public static BigDecimal newtonRaphson0(
            BigDecimal x,
            BigDecimal result0,
            BigInteger root_BigInteger,
            BigDecimal epsilon_BigDecimal,
            int maxite,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        BigDecimal result = new BigDecimal(result0.toString());
        int safeDecimalPlaces = decimalPlaces + 5;
        BigInteger rootsubtract1_BigInteger = root_BigInteger.subtract(
                BigInteger.ONE);
        BigDecimal previousResult_BigDecimal;
        BigDecimal previousResult_BigDecimal0;
        BigDecimal resultipowroot;
        BigDecimal resultipowrootsubtract1;
        BigDecimal divisor;
        BigDecimal resultipowrootsubtractx;
        BigDecimal resultipowrootsubtractxdividedivisor;
        boolean maxiteReached = true;
        // Initialise toCompare and previousResult_BigDecimal
        previousResult_BigDecimal = new BigDecimal(result0.toString());
        previousResult_BigDecimal0 = new BigDecimal(result0.toString());
        for (int i = 0; i < maxite; i ++) {
            resultipowroot = power(
                    result,
                    root_BigInteger,
                    64,
                    safeDecimalPlaces,
                    a_RoundingMode);
            resultipowrootsubtract1 = power(
                    result,
                    rootsubtract1_BigInteger,
                    64,
                    safeDecimalPlaces,
                    a_RoundingMode);
            divisor = multiply(
                    resultipowrootsubtract1,
                    root_BigInteger,
                    safeDecimalPlaces,
                    a_RoundingMode);
            resultipowrootsubtractx = resultipowroot.subtract(x);
            resultipowrootsubtractxdividedivisor = divide(
                    resultipowrootsubtractx,
                    divisor,
                    safeDecimalPlaces,
                    a_RoundingMode);
            result = result.subtract(resultipowrootsubtractxdividedivisor);
            // Test for convergence
            if (previousResult_BigDecimal.subtract(result).abs().compareTo(epsilon_BigDecimal) != 1) {
                //System.out.println("Root found in iteration " + i + " out of " + maxite);
                maxiteReached = false;
                break;
            }
            previousResult_BigDecimal0 = new BigDecimal(previousResult_BigDecimal.toString());
            previousResult_BigDecimal = new BigDecimal(result.toString());
        }
        if (maxiteReached) {
            System.out.println("previousResult_BigDecimal0.subtract(result).abs() " + previousResult_BigDecimal0.subtract(result).abs());
        }
        return result;
    }

//    public static BigDecimal newtonRaphson(
//            BigDecimal x,
//            BigDecimal result0,
//            BigInteger root,
//            BigDecimal epsilon_BigDecimal,
//            int maxite,
//            int decimalPlaces,
//            RoundingMode a_RoundingMode) {
//        BigDecimal result = new BigDecimal(result0.toString());
//        int safeDecimalPlaces = decimalPlaces + 5;
//        BigInteger rootsubtract1_BigDecimal = root.subtract(BigInteger.ONE);
//        BigDecimal previousResult_BigDecimal;
//        BigDecimal previousResult_BigDecimal0;
//        BigDecimal resultipowroot;
//        BigDecimal resultipowrootsubtract1;
//        BigDecimal divisor;
//        BigDecimal resultipowrootsubtractx;
//        BigDecimal resultipowrootsubtractxdividedivisor;
//        boolean maxiteReached = true;
//        // Initialise toCompare and previousResult_BigDecimal
//        previousResult_BigDecimal = new BigDecimal("1");
//        previousResult_BigDecimal0 = new BigDecimal("1");
//        int i = 0;
//        while (i < maxite) {
//            resultipowroot = power(
//                    result,
//                    root,
//                    64,
//                    decimalPlaces,
//                    a_RoundingMode);
//            resultipowrootsubtract1 = power(
//                    result,
//                    rootsubtract1_BigDecimal,
//                    64,
//                    decimalPlaces,
//                    a_RoundingMode);
//            divisor = multiply(
//                    resultipowrootsubtract1,
//                    root,
//                    safeDecimalPlaces,
//                    a_RoundingMode);
//            resultipowrootsubtractx = resultipowroot.subtract(x);
//            resultipowrootsubtractxdividedivisor = divide(
//                    resultipowrootsubtractx,
//                    divisor,
//                    safeDecimalPlaces,
//                    a_RoundingMode);
//            result = result.subtract(resultipowrootsubtractxdividedivisor);
//            // Test for convergence
//            if (previousResult_BigDecimal.subtract(result).abs().compareTo(epsilon_BigDecimal) != 1) {
//                System.out.println("Root found in iteration " + i + " out of " + maxite);
//                maxiteReached = false;
//                break;
//            }
//            previousResult_BigDecimal0 = new BigDecimal(previousResult_BigDecimal.toString());
//            previousResult_BigDecimal = new BigDecimal(result.toString());
//            i++;
//        }
//        if (maxiteReached) {
//            System.out.println("previousResult_BigDecimal0.subtract(result).abs() " + previousResult_BigDecimal0.subtract(result).abs());
//        }
//        return result;
//    }
    private static BigDecimal rootInitialisation(
            BigDecimal x,
            BigInteger root,
            BigDecimal epsilon_BigDecimal,
            int maxite,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        BigDecimal result;
        int div = 64;
        // Initialise toCompare and previousResult_BigDecimal
        if (x.compareTo(BigDecimal.ONE) == -1) {
            return rootInitialisationLessThanOne(
                    x,
                    root,
                    epsilon_BigDecimal,
                    maxite,
                    decimalPlaces,
                    a_RoundingMode);
        } else {
            int i = 0;
            result = BigDecimal.ONE.add(epsilon_BigDecimal);
            boolean powerTestAbove = powerTestAbove(
                    x,
                    result,
                    root,
                    64,
                    decimalPlaces,
                    a_RoundingMode);
            if (powerTestAbove) {
                // Root cannot be found within current precision...
                int debug = 1;
                return BigDecimal.ONE;
            }
            BigDecimal a = new BigDecimal(x.toString());
            BigDecimal b = BigDecimal.ONE;
            BigDecimal c;

            BigInteger rootdiv = new BigInteger(root.toString());
            //boolean notInitialised = true;
            //BigDecimal difference_BigDecimal;
//            while (i < maxite && notInitialised) {
            while (i < maxite) {
//                difference_BigDecimal = a.subtract(b);
//                if (difference_BigDecimal.compareTo(epsilon_BigDecimal) == -1) {
//                    notInitialised = false;
//                }
//                if (notInitialised) {
                // Disect
                c = divide(
                        a.subtract(b),
                        rootdiv,
                        decimalPlaces + 1,
                        a_RoundingMode);
                c = b.add(c);
                powerTestAbove = powerTestAbove(
                        x,
                        c,
                        root,
                        div,
                        decimalPlaces,
                        a_RoundingMode);
                if (powerTestAbove) {
                    a = c;
                    //a = new BigDecimal(c.toString());
                } else {
                    b = c;
                    //b = new BigDecimal(c.toString());
                    rootdiv = rootdiv.divide(Generic_BigInteger.Two);
                    result = b;
//                        difference_BigDecimal = (result.subtract(c)).abs();
//                        if (difference_BigDecimal.compareTo(epsilon_BigDecimal) == -1) {
//                            notInitialised = false;
//                        }
                }
//                if (notInitialised) {
                // Bisect
                c = divide(
                        a.subtract(b),
                        Two,
                        decimalPlaces + 1,
                        a_RoundingMode);
                c = c.add(b);
                powerTestAbove = powerTestAbove(
                        x,
                        c,
                        root,
                        div,
                        decimalPlaces,
                        a_RoundingMode);
                if (powerTestAbove) {
                    a = c;
                    //a = new BigDecimal(c.toString());
                } else {
                    b = c;
                    //b = new BigDecimal(c.toString());
                    result = b;
//                            difference_BigDecimal = (result.subtract(c)).abs();
//                            if (difference_BigDecimal.compareTo(epsilon_BigDecimal) == -1) {
//                                notInitialised = false;
//                            }
                }
//                }
//                }
                i++;
            }
        }
        return result;
    }

//    private static BigDecimal rootInitialisation0(
//            BigDecimal x,
//            BigDecimal root_BigDecimal,
//            BigDecimal maxError_BigDecimal,
//            BigDecimal a0,
//            BigDecimal e,
//            Generic_BigDecimal a_Generic_BigDecimal,
//            int decimalPlaces,
//            MathContext a_MathContext,
//            RoundingMode a_RoundingMode) {
//        BigDecimal result;
//        if (x.compareTo(BigDecimal.ONE) == -1) {
//            result = BigDecimal.ONE.subtract(maxError_BigDecimal);
//            // Bisect
//            BigDecimal a = a0;
//            BigDecimal b = result;
//            BigDecimal c;
//            BigDecimal cPowerRoot;
//            boolean notInitialised = true;
//            while (notInitialised) {
//                c = (a.add(b)).divide(Two, decimalPlaces, a_RoundingMode);
//                cPowerRoot = power(
//                        c,
//                        root_BigDecimal,
//                        e,
//                        a_Generic_BigDecimal,
//                        decimalPlaces,
//                        a_RoundingMode);
//                if (cPowerRoot.compareTo(x) == 1) {
//                    b = c;
//                    if (a.compareTo(c) == 0) {
//                        notInitialised = false;
//                    }
//                } else {
//                    result = c;
//                    notInitialised = false;
//                }
//            }
//        } else {
//            result = BigDecimal.ONE.add(maxError_BigDecimal);
//            BigDecimal a = x.divide(root_BigDecimal, decimalPlaces, a_RoundingMode);
//            if (a.compareTo(BigDecimal.ONE) == 1) {
//                result = BigDecimal.ONE.add(maxError_BigDecimal);
//                // Disect
//                BigDecimal b = a0;
//                BigDecimal c;
//                boolean notInitialised = true;
//                BigDecimal cPowerRoot;
//                while (notInitialised) {
//                    c = (a.add(b)).divide(root_BigDecimal, decimalPlaces, a_RoundingMode);
//                    cPowerRoot = power(
//                            c,
//                            root_BigDecimal,
//                            e,
//                            a_Generic_BigDecimal,
//                            decimalPlaces,
//                            a_RoundingMode);
//                    if (cPowerRoot.compareTo(x) == 1) {
//                        a = c;
//                    } else {
//                        // Bisect
//                        c = (a.add(b)).divide(Two, decimalPlaces, a_RoundingMode);
//                        cPowerRoot = power(
//                                c,
//                                root_BigDecimal,
//                                e,
//                                a_Generic_BigDecimal,
//                                decimalPlaces,
//                                a_RoundingMode);
//                        if (cPowerRoot.compareTo(x) == 1) {
//                            a = c;
//                        } else {
//                            result = c;
//                            notInitialised = false;
//                        }
//                    }
//                }
//            } else {
//                // Bisect
//                a = x.divide(Two, decimalPlaces, a_RoundingMode);
//                BigDecimal b = result;
//                BigDecimal c;
//                BigDecimal aAddb;
//                BigDecimal aAddbSubtractTwo;
//                BigDecimal cPowerroot;
//                boolean notInitialised = true;
//                BigDecimal cPowerRoot;
//                while (notInitialised) {
//                    // Disect
//                    aAddb = a.add(b);
//                    aAddbSubtractTwo = aAddb.subtract(Two);
//                    if (aAddbSubtractTwo.compareTo(BigDecimal.ZERO) != 1) {
//                        notInitialised = false;
//                    } else {
//                        c = (aAddb).divide(aAddbSubtractTwo, decimalPlaces, a_RoundingMode);
//                        cPowerRoot = power(
//                                c,
//                                root_BigDecimal,
//                                e,
//                                a_Generic_BigDecimal,
//                                decimalPlaces,
//                                a_RoundingMode);
//                        if (cPowerRoot.compareTo(x) == 1) {
//                            a = c;
//                        } else {
//                            while (notInitialised) {
//                                // Bisect
//                                a = a.divide(Two, decimalPlaces, a_RoundingMode);
//                                b = c;
//                                aAddb = a.add(b);
//                                aAddbSubtractTwo = aAddb.subtract(BigDecimal.ONE);
//                                c = (aAddb).divide(aAddbSubtractTwo, decimalPlaces, a_RoundingMode);
//                                cPowerRoot = power(
//                                        c,
//                                        root_BigDecimal,
//                                        e,
//                                        a_Generic_BigDecimal,
//                                        decimalPlaces,
//                                        a_RoundingMode);
//                                if (cPowerRoot.compareTo(x) == 1) {
//                                    a = c;
//                                } else {
//                                    result = c;
//                                    notInitialised = false;
//                                }
//                            }
//                        }
//                    }
//                }
//            }
//        }
//        return result;
//    }
    /**
     * @param x x >= 0
     * @param root
     * @param decimalPlaces
     * @param a_RoundingMode
     * @return The root root of x to decimalPlaces precision.
     */
    public static BigDecimal root(
            BigDecimal x,
            int root,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        // Deal with special cases
        if (x.compareTo(BigDecimal.ZERO) == -1) {
            throw new IllegalArgumentException(
                    "x < 0 in " + Generic_BigDecimal.class
                    + ".root(BigDecimal,int,int,RoundingMode)");
        }
        if (x.compareTo(BigDecimal.ZERO) == 0) {
            return BigDecimal.ZERO;
        }
        if (x.compareTo(BigDecimal.ONE) == 0) {
            return BigDecimal.ONE;
        }
        if (root == 0) {
            throw new IllegalArgumentException(
                    "root = 0 in " + Generic_BigDecimal.class
                    + ".root(BigDecimal,int,int,RoundingMode)");
        }
        if (root == 1) {
            return new BigDecimal(x.toString());
        }
        // The current (Java6) limit for n in x.pow(n)
        // for BigDecimal x and int n is
        // -999999999 < n < 999999999
        if (root >= 999999999) {
            Generic_BigDecimal a_Generic_BigDecimal =
                    new Generic_BigDecimal(root);
            return root(
                    x,
                    BigInteger.valueOf(root),
                    decimalPlaces,
                    a_RoundingMode);
        }
        if (x.compareTo(BigDecimal.ONE) == -1) {
            return rootLessThanOne(
                    x,
                    root,
                    decimalPlaces,
                    a_RoundingMode);
        }
        BigDecimal epsilon_BigDecimal =
                new BigDecimal(BigInteger.ONE, decimalPlaces);
        BigDecimal comparator = BigDecimal.ONE.add(epsilon_BigDecimal);
        // Check there is a root in the precision and return 0 if not
        BigInteger root_BigInteger = BigInteger.valueOf(root);
        boolean powerTest = powerTestAbove(
                x,
                comparator,
                root_BigInteger,
                256,
                decimalPlaces,
                a_RoundingMode);
        if (powerTest) {
            System.out.println("No root in the precision... ");
            return BigDecimal.ZERO;
        }
        BigDecimal result = BigDecimal.ONE;
        int rootInitialisationMaxite = 10;
        result = rootInitialisation(
                x,
                root_BigInteger,
                epsilon_BigDecimal,
                rootInitialisationMaxite,
                decimalPlaces,
                a_RoundingMode);
        int newtonRaphsonMaxite = 100;
        result = newtonRaphson(
                x,
                result,
                root_BigInteger,
                epsilon_BigDecimal,
                newtonRaphsonMaxite,
                decimalPlaces,
                a_RoundingMode);
        return result;
    }

    /**
     * @param x 0 < x < 1
     * @param root
     * @param decimalPlaces
     * @param a_RoundingMode
     * @return The root root of x to decimalPlaces precision.
     */
    private static BigDecimal rootLessThanOne(
            BigDecimal x,
            int root,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        return rootLessThanOne(
                x,
                BigInteger.valueOf(root),
                decimalPlaces,
                a_RoundingMode);
    }

    /**
     * @param x 0 < x < 1
     * @param root
     * @param decimalPlaces
     * @param a_RoundingMode
     * @return The root root of x to decimalPlaces precision.
     */
    private static BigDecimal rootLessThanOne(
            BigDecimal x,
            BigInteger root,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        BigDecimal epsilon_BigDecimal =
                new BigDecimal(BigInteger.ONE, decimalPlaces);
        //BigDecimal comparator = BigDecimal.ONE.subtract(epsilon_BigDecimal);
//        // Check there is a root in the precision and return 1 if not
//        boolean powerTest = powerTestBelow(
//                epsilon_BigDecimal,//comparator,
//                x,
//                root,
//                256,
//                decimalPlaces,
//                a_RoundingMode);
//        if (powerTest) {
//            System.out.println("No root in the precision... ");
//            return BigDecimal.ONE;
//        }
        BigDecimal result = BigDecimal.ONE;
        int rootInitialisationMaxite = 10;
        result = rootInitialisationLessThanOne(
                x,
                root,
                epsilon_BigDecimal,
                rootInitialisationMaxite,
                decimalPlaces,
                a_RoundingMode);
        int newtonRaphsonMaxite = 100;
        return newtonRaphson(
                x,
                result,
                root,
                epsilon_BigDecimal,
                newtonRaphsonMaxite,
                decimalPlaces,
                a_RoundingMode);
    }

//    /**
//     * @return true if there is a root in the precision and false otherwise
//     */
//    private boolean isRootInPrecision(
//            BigDecimal x,
//            int root,
//            BigDecimal epsilon_BigDecimal,
//            int precision,
//            RoundingMode a_RoundingMode){
//        boolean toCompare = true;
//            if (x.compareTo(BigDecimal.ONE) == 1) {
//                BigDecimal check = power0(
//                        epsilon_BigDecimal,
//                        root,
//                        root,
//                        4,
//                        precision,
//                        a_RoundingMode);
////                BigDecimal check = power0( // Too slow...
////                        epsilon_BigDecimal,
////                        root,
////                        precision,
////                        a_RoundingMode);
//                if (check.compareTo(x) == 1) {
//                    System.out.println("No root in the precision...");
//                    return BigDecimal.ZERO;
//                }
//            }
//    }
//    private static BigDecimal rootInitialisation(
//            BigDecimal x,
//            int root,
//            //BigDecimal root_BigDecimal,
//            BigDecimal maxError_BigDecimal,
//            BigDecimal a0,
//            int decimalPlaces,
//            RoundingMode a_RoundingMode) {
//        BigDecimal result;
//        // Initialise toCompare and previousResult_BigDecimal
//        if (x.compareTo(BigDecimal.ONE) == -1) {
//            return rootInitialisationLessThanOne(
//                    x,
//                    root,
//                    //root_BigDecimal,
//                    maxError_BigDecimal,
//                    a0,
//                    decimalPlaces,
//                    a_RoundingMode);
//        } else {
//            result = BigDecimal.ONE.add(maxError_BigDecimal);
//            BigDecimal a = x.divide(root_BigDecimal, decimalPlaces, a_RoundingMode);
//            if (a.compareTo(BigDecimal.ONE) == 1) {
//                result = BigDecimal.ONE.add(maxError_BigDecimal);
//                // Disect
//                BigDecimal b = a0;
//                BigDecimal c;
//                //BigDecimal cPowerRoot;
//                boolean cPowerRootTest;
//                boolean notInitialised = true;
//                while (notInitialised) {
//                    c = (a.add(b)).divide(root_BigDecimal, decimalPlaces, a_RoundingMode);
//                    cPowerRootTest = powerTestAbove(
//                            x,
//                            c,
//                            root,
//                            64,
//                            decimalPlaces,
//                            a_RoundingMode);
//                    if (cPowerRootTest) {
////                    cpowerroot = power(
////                            c,
////                            root,
////                            64,
////                            decimalPlaces,
////                            a_RoundingMode);
////                    if (cpowerroot.compareTo(x) == 1) {
//                        a = c;
//                    } else {
//                        // Bisect
//                        c = (a.add(b)).divide(Two, decimalPlaces, a_RoundingMode);
//                        cPowerRootTest = powerTestAbove(
//                                x,
//                                c,
//                                root,
//                                64,
//                                decimalPlaces,
//                                a_RoundingMode);
//                        if (cPowerRootTest) {
////                        cpowerroot = power(
////                            c,
////                            root,
////                            64,
////                            decimalPlaces,
////                            a_RoundingMode);
////                        if (cpowerroot.compareTo(x) == 1) {
//                            a = c;
//                        } else {
//                            result = c;
//                            notInitialised = false;
//                        }
//                    }
//                }
//            } else {
//                boolean resultpowerroottest = powerTestAbove(
//                        x,
//                        result,
//                        root,
//                        64,
//                        decimalPlaces,
//                        a_RoundingMode);
//                if (resultpowerroottest) {
//                    // Root cannot be found within current precision...
//                    // return BigDecimal.ZERO or BigDecimal.ONE?
//                    int debug = 1;
//                    return BigDecimal.ONE;
//                }
////                BigDecimal resultpowroot = power(
////                        result,
////                        root,
////                        64,
////                        decimalPlaces,
////                        a_RoundingMode);
////                if (resultpowroot.compareTo(x) == 1) {
////                    // Root cannot be found within current precision...
////                    // return BigDecimal.ZERO or BigDecimal.ONE?
////                    int debug = 1;
////                }
//                a = x.divide(Two, decimalPlaces, a_RoundingMode);
//                BigDecimal b = new BigDecimal(result.toString());
//                BigDecimal c;
//                //BigDecimal cPowerRoot;
//                boolean cPowerRootTest;
//                BigDecimal aAddb;
//                boolean notInitialised = true;
//                int maxite = 1000;
//                int i = 0;
//                while (notInitialised) {
//                    if (i == maxite) {
//                        notInitialised = false;
//                    }
//                    // Bisect
//                    aAddb = a.add(b);
//                    c = (aAddb).divide(Two, decimalPlaces, a_RoundingMode);
//                    cPowerRootTest = powerTestAbove(
//                            x,
//                            c,
//                            root,
//                            64,
//                            decimalPlaces,
//                            a_RoundingMode);
//                    if (cPowerRootTest) {
////                    cpowerroot = power(
////                            c,
////                            root,
////                            64,
////                            decimalPlaces,
////                            a_RoundingMode);
////                    if (cpowerroot.compareTo(x) == 1) {
//                        a = c;
//                    } else {
//                        //result = c;
//                        notInitialised = false;
//                    }
//                    result = a;
//                    i++;
//                }
//            }
//        }
//        return result;
//    }
    /**
     *
     * @param x 0 < x < 1
     * @param root_BigDecimal
     * @param maxError_BigDecimal
     * @param a0
     * @param decimalPlaces
     * @param a_MathContext
     * @param a_RoundingMode
     * @return
     */
    private static BigDecimal rootInitialisationLessThanOne(
            BigDecimal x,
            BigInteger root,
            BigDecimal epsilon_BigDecimal,
            int maxite,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        BigDecimal b = BigDecimal.ONE;
        BigDecimal result = BigDecimal.ONE.subtract(epsilon_BigDecimal);
        boolean powerTestBelow;
//        boolean powerTestBelow = powerTestBelow(
//                x,
//                result,
//                root,
//                64,
//                decimalPlaces,
//                a_RoundingMode);
//        if (powerTestBelow) {
//            // No root in precision
//            int debug = 1;
//            return b;
//        }
        BigDecimal a = new BigDecimal(x.toString());
        b = result;
        BigDecimal c;
//        boolean notInitialised = true;
        BigInteger rootdiv = new BigInteger(root.toString(maxite));
        int i = 0;
        while (i < maxite) {
            // Disect;
            c = divide(
                    b.subtract(a),
                    root,
                    decimalPlaces + 1,
                    a_RoundingMode);
            c = b.subtract(c);
            powerTestBelow = powerTestBelow(
                    x,
                    c,
                    root,
                    64,
                    decimalPlaces,
                    a_RoundingMode);
            if (powerTestBelow) {
                a = c;
            } else {
                b = c;
                rootdiv = rootdiv.divide(Generic_BigInteger.Two);
                result = b;
            }
            // Bisect;
            c = divide(
                    b.subtract(a),
                    Two,
                    decimalPlaces + 1,
                    a_RoundingMode);
            c = b.subtract(c);
            powerTestBelow = powerTestBelow(
                    x,
                    c,
                    root,
                    64,
                    decimalPlaces,
                    a_RoundingMode);
            if (powerTestBelow) {
                a = c;
            } else {
                b = c;
                result = b;
            }
            i++;
        }
        return result;
    }

    /**
     * Returns the square root of the input BigDecimal as a BigDecimal. This can
     * only handle input with a precision of less than Integer.MAX_VALUE. No
     * checking for this is done.
     * @param input
     * @param decimalPlaces
     * @return Square root of input
     */
    public static BigDecimal sqrt0(
            BigDecimal input,
            int decimalPlaces) {
        // Deal with special cases
        BigDecimal zero = new BigDecimal(BigInteger.ZERO);
        if (input.compareTo(zero) == -1) {
            return null;
        }
        if (input.compareTo(BigDecimal.ONE) == 0) {
            return BigDecimal.ONE;
        } // Deal with general case ( input > 1 or 0 > input < 1 )
        // Calculate number of digits numberOfDigits in toCompare
        BigInteger b0 = input.toBigInteger();
//        System.out.println( b0.toString() );
        int numberOfDigits = b0.toString().length() / 2;
        BigInteger b1;
        if (numberOfDigits == 0) {
            b1 = BigInteger.ONE;
        } else {
            b1 = new BigInteger("10").pow(numberOfDigits - 1);
        }
        // Calculate value of first digit
        int i = 1;
        BigInteger b2 = b1;
        BigInteger b3 = b2.multiply(new BigInteger(Integer.toString(i)));
        BigInteger b4 = null;
        while (b0.compareTo(b2) == 1) {
            b4 = b3;
            b3 = b1.multiply(new BigInteger(Integer.toString(i)));
            b2 = (b3).pow(2);
            i++;
        }
        if (b4 == null) {
            b4 = BigInteger.ONE;
        }
        //System.out.println( "square root of " + input.toString() + " is between " + b4 + " and " + b3.toString() );
        //System.out.println( b4 + " squared is " + b4.pow( 2 ) );
        //System.out.println( b3 + " squared is " + b3.pow( 2 ) );
        //System.out.println( "Done digit 0" );
        // Calculate value of remaining digits that are greater than 1
        for (int ite = 1; ite < numberOfDigits; ite++) {
            b1 = new BigInteger("10").pow(numberOfDigits - 1 - ite);
            i = 1;
            b2 = b4;
            b3 = b2.add(b2.multiply(b1));
            while (b0.compareTo(b2) == 1) {
                b4 = b3;
                b3 = b1.multiply(new BigInteger(Integer.toString(i)));
                b2 = (b3).pow(2);
                i++;
            }
            //System.out.println( "square root of " + input.toString() + " is between " + b4.toString() + " and " + b3.toString() );
            //System.out.println( b4.toString() + " squared is " + b4.pow( 2 ) );
            //System.out.println( b3.toString() + " squared is " + b3.pow( 2 ) );
            //System.out.println( "Done digit " + ite );
        }
        // Iteratively approximate fractions until required precision (decimal places) is reached
        int roundingMode = BigDecimal.ROUND_HALF_DOWN;
        BigDecimal half = new BigDecimal(0.5d);
//        System.out.println("" + half.toString() );
        BigDecimal x = new BigDecimal(b4);
        BigDecimal y;
        int iteScale = decimalPlaces + 1;
        int divideScale = decimalPlaces + 10;
        do {//for ( int ite = 0; ite < 10; ite ++ ) {
            //System.out.println("" + input.toString() + " divide " + toCompare.toString() + " equals " + input.divide( toCompare, scale, roundingMode ).toString() );
            y = half.multiply((x.add((input.divide(x, divideScale, roundingMode)))));
            x = y;
            // System.out.println( "square root of " + input.toString() + " is nearly " + toCompare.toString() );
//            System.out.println( toCompare.toString() + " squared is " + toCompare.multiply( toCompare ) );
            // System.out.println( "" + ( toCompare.multiply( toCompare ) ).setScale( scale, roundingMode ) );
        } while (((x.multiply(x)).setScale(decimalPlaces, roundingMode)).compareTo(input.setScale(decimalPlaces, roundingMode)) != 0);
        roundingMode = BigDecimal.ROUND_DOWN;
        y = x.setScale(decimalPlaces, roundingMode);
        //System.out.println( y + " squared is " + y.multiply( y ) );
        BigDecimal smallestDecimalFraction = new BigDecimal(BigInteger.ONE, decimalPlaces);
        // Choose closest to return
        BigDecimal yplus = y.add(smallestDecimalFraction);
        //System.out.println( yplus + " squared is " + yplus.multiply( yplus ) );
        //BigDecimal yminus = y.subtract( smallestDecimalFraction );
        BigDecimal diffplus = (yplus.multiply(yplus)).subtract(input);
        BigDecimal diff = (y.multiply(y)).subtract(input);
        //BigDecimal diffminus = ( yplus.multiply( yplus ) ).subtract( input );
        if (diffplus.compareTo(diff) == -1) {
            //if ( diffplus.compareTo( diffminus ) == -1 ) {
            //System.out.println( "returning " + yplus );
            return yplus;
            //} else {
            //    return yminus;
            //}
        } else {
            //if ( diff.compareTo( diffminus ) == -1 ) {
            //System.out.println( "returning " + y );
            return y;
            //} else {
            //    return yminus;
            //}
        }
    }
//    /**
//     * Returns the square root of the input BigDecimal as a BigDecimal. There
//     * are limits to the size of BigDecimals this can handle, but they can have
//     * Integer.MAX_VALUE numbers of digits in their decimal expansion and so
//     * can be enormous! An exception should be thrown and caught if this is the
//     * case but this needs implementing.
//     */
//    public static BigDecimal sqrt1( BigDecimal input, int decimalPlaces ) {
//        System.out.println( "sqrt(" + input + "," + decimalPlaces + ") {");
//        BigDecimal toCompare;
//        int inputScale = input.scale();
//        BigInteger bi0 = input.unscaledValue();
//        System.out.println( "bi0 " + bi0 );
//        int numberOfDigits = ( bi0.toString().length() - 1 ) / 2;
//        BigInteger bi1 = new BigInteger( "10" ).pow( numberOfDigits );
//
//        System.out.println( "bi1 " + bi1 );
//        int i = 1;
//        BigInteger bi2 = bi1;
//        BigInteger bi3 = bi2.multiply( new BigInteger( Integer.toString( i ) ) );
//        BigInteger bi4 = null;
//        while ( bi0.compareTo( bi2 ) == 1 ) {
//            bi4 = bi3;
//            bi3 = bi1.multiply( new BigInteger( Integer.toString( i ) ) );
//            bi2 = ( bi3 ).pow( 2 );
//            i ++;
//        }
//        System.out.println( "square root of " + bi0.toString() + " is between " + bi4.toString() + " and " + bi3.toString() );
//        System.out.println( bi4.toString() + " squared is " + bi4.pow( 2 ) );
//        System.out.println( bi3.toString() + " squared is " + bi3.pow( 2 ) );
//        System.out.println( "Done 0" );
//
//        for ( int ite = 1; ite < numberOfDigits + 1; ite ++ ) {
//            bi1 = new BigInteger( "10" ).pow( numberOfDigits - ite );
//            i = 1;
//            bi2 = bi4;
//            bi3 = bi2.add( bi2.multiply( bi1 ) );
//            while ( bi0.compareTo( bi2 ) == 1 ) {
//                bi4 = bi3;
//                bi3 = bi1.multiply( new BigInteger( Integer.toString( i ) ) );
//                bi2 = ( bi3 ).pow( 2 );
//                i ++;
//            }
//            System.out.println( "square root of " + bi0.toString() + " is between " + bi4.toString() + " and " + bi3.toString() );
//            System.out.println( bi4.toString() + " squared is " + bi4.pow( 2 ) );
//            System.out.println( bi3.toString() + " squared is " + bi3.pow( 2 ) );
//            System.out.println( "Done " + ite );
//        }
//
//        // y = (1/2)(toCompare + a/toCompare).
//        int roundingMode = BigDecimal.ROUND_HALF_EVEN;
//        //BigDecimal one = new BigDecimal( BigInteger.ONE );
//        //BigDecimal two = new BigDecimal( new BigInteger( "2" ) );
//        //BigDecimal half = one.divide( two, scale, roundingMode );
//        BigDecimal half = new BigDecimal( 0.5d );
//        System.out.println("" + half.toString() );
//        BigDecimal toCompare = new BigDecimal( bi4 );
//        BigDecimal y;
//        int divideScale = inputScale * 2;
//
//        do {//for ( int ite = 0; ite < 10; ite ++ ) {
//            //System.out.println("" + input.toString() + " divide " + toCompare.toString() + " equals " + input.divide( toCompare, scale, roundingMode ).toString() );
//            y = half.multiply( ( toCompare.add( ( input.divide( toCompare, divideScale, roundingMode ) ) ) ) );
//            toCompare = y;
//            // System.out.println( "square root of " + input.toString() + " is nearly " + toCompare.toString() );
//            // System.out.println( toCompare.toString() + " squared is " + toCompare.multiply( toCompare ) );
//            // System.out.println( "" + ( toCompare.multiply( toCompare ) ).setScale( scale, roundingMode ) );
//            //} while ( ( ( toCompare.multiply( toCompare ) ).setScale( scale, roundingMode ) ).compareTo( input ) != 0 );
//        } while ( ( ( toCompare.multiply( toCompare ) ).setScale( inputScale, roundingMode ) ).compareTo( input ) != 0 );
//        y = toCompare.setScale( decimalPlaces, roundingMode );
//        System.out.println( y.toString() + " squared is " + y.multiply( y ) );
//
//        BigDecimal check = new BigDecimal( BigInteger.ONE, decimalPlaces );
//        System.out.println( y.add(check).toString() + " squared is " + y.add(check).multiply( y.add(check) ) );
//        System.out.println( y.subtract(check).toString() + " squared is " + y.subtract(check).multiply( y.subtract(check) ) );
//        //System.out.println( "" + y );
//        return y;
//    }
}
