/**
 * Copyright 2010 Andy Turner, The University of Leeds, UK
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package uk.ac.leeds.ccg.andyt.generic.io;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.StreamTokenizer;
import java.io.BufferedOutputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Map.Entry;
import java.util.TreeMap;
import uk.ac.leeds.ccg.andyt.generic.utilities.Generic_ErrorAndExceptionHandler;
//import java.nio.file.S

//TODO http://java.sun.com/docs/books/tutorial/essential/io/legacy.html#mapping
// http://java.sun.com/docs/books/tutorial/essential/io/fileio.html
/**
 * Class of static methods for helping with reading and writing (primarily)
 * to/from file.
 */
public class Generic_StaticIO {

    /** Creates a new instance of ObjectIO */
    public Generic_StaticIO() {
    }

    /**
     * Write object to file
     */
    public static void writeObject(
            Object object,
            File file) {
        try {
            ObjectOutputStream _ObjectOutputStream =
                    new ObjectOutputStream(
                    new BufferedOutputStream(
                    new FileOutputStream(file)));
            _ObjectOutputStream.writeObject(object);
            _ObjectOutputStream.flush();
            _ObjectOutputStream.close();
        } catch (IOException aIOException) {
            aIOException.printStackTrace();
            System.exit(Generic_ErrorAndExceptionHandler.IOException);
        }
    }

    /**
     * Read Object from File
     */
    public static Object readObject(
            File file) {
        Object result = null;
        if (file.length() != 0) {
            try {
                ObjectInputStream _ObjectInputStream =
                        new ObjectInputStream(
                        new BufferedInputStream(
                        new FileInputStream(file)));
                result = _ObjectInputStream.readObject();
                _ObjectInputStream.close();
            } catch (IOException aIOException) {
                aIOException.printStackTrace();
                System.exit(Generic_ErrorAndExceptionHandler.IOException);
            } catch (ClassNotFoundException aClassNotFoundException) {
                aClassNotFoundException.printStackTrace();
                System.exit(Generic_ErrorAndExceptionHandler.ClassNotFoundException);
            }
        }
        return result;
    }

    /**
     * Attempts to make a copy of input_File in _Output_Directory or as a file
     * in it if it is a File rather than a directory.
     * @param input_File A File which is not a Directory to be copied
     * @param outputDirectory_File A File that is the Directory to copy to.
     */
    public static void copyFile(
            File input_File,
            File outputDirectory_File) {
        // String osName = System.getProperty( "os.name" );
        // if ( osName.equalsIgnoreCase( "UNIX" ) ) {
        // Runtime.getRuntime().exec( "cp " + _InputFile.toString() + " " +
        // _Output_File.toString() );
        // } else {
        // }
        if (!input_File.exists()) {
            System.err.println(
                    "!input_File.exists() in "
                    + Generic_StaticIO.class.getCanonicalName()
                    + ".copy(File(" + input_File + "),File(" + outputDirectory_File + "))");
            System.exit(Generic_ErrorAndExceptionHandler.IOException);
        }
        if (!outputDirectory_File.exists()) {
            outputDirectory_File.mkdirs();
        }
        File output_File = new File(
                outputDirectory_File, input_File.getName());
        if (output_File.exists()) {
            System.out.println(
                    "Overwriting File " + output_File + " in "
                    + Generic_StaticIO.class.getCanonicalName()
                    + ".copy(File(" + input_File + "),"
                    + "File(" + outputDirectory_File + "))");
        } else {
            try {
                output_File.createNewFile();
            } catch (IOException aIOException) {
                aIOException.printStackTrace();
                System.err.println(
                        "Unable to createNewFile " + output_File + " in "
                        + Generic_StaticIO.class.getCanonicalName()
                        + ".copy(File(" + input_File + "),"
                        + "File(" + outputDirectory_File + "))");
                System.exit(Generic_ErrorAndExceptionHandler.IOException);
            }
        }
        try {
            BufferedInputStream bis = new BufferedInputStream(
                    new FileInputStream(input_File));
            BufferedOutputStream bos = new BufferedOutputStream(
                    new FileOutputStream(output_File));
            // bufferSize should be power of 2 (e.g. Math.pow(2, 12)), but nothing too big.
            int bufferSize = 2048;
            long numberOfArrayReads = input_File.length() / bufferSize;
            long numberOfSingleReads = input_File.length() - (numberOfArrayReads * bufferSize);
            byte[] b = new byte[bufferSize];
            for (int i = 0; i < numberOfArrayReads; i++) {
                bis.read(b);
                bos.write(b);
            }
            for (int i = 0; i < numberOfSingleReads; i++) {
                bos.write(bis.read());
            }
            bos.flush();
            bos.close();
            bis.close();
        } catch (IOException aIOException) {
            aIOException.printStackTrace();
            System.exit(Generic_ErrorAndExceptionHandler.IOException);
        }
    }

    public static void copyDirectory(
            File a_DirectoryToCopy_File,
            File a_DirectoryToCopyInto_File) {
        try {
            if (!a_DirectoryToCopyInto_File.mkdir()) {
                a_DirectoryToCopyInto_File.mkdirs();
            }
            if (!a_DirectoryToCopy_File.isDirectory()) {
                throw new IOException(
                        "Expecting File " + a_DirectoryToCopy_File
                        + "To be a directory in "
                        + Generic_StaticIO.class.getName()
                        + ".copyDirectory(File,File)");
            }
            if (!a_DirectoryToCopyInto_File.isDirectory()) {
                throw new IOException(
                        "Expecting File " + a_DirectoryToCopyInto_File
                        + "To be a directory in "
                        + Generic_StaticIO.class.getName() + ".copy(File,File)");
            }
            File copiedDirectory_File = new File(
                    a_DirectoryToCopyInto_File,
                    a_DirectoryToCopy_File.getName());
            copiedDirectory_File.mkdir();
            File[] a_DirectoryToCopy_File_Files =
                    a_DirectoryToCopy_File.listFiles();
            for (int i = 0; i < a_DirectoryToCopy_File_Files.length; i ++){
                if (a_DirectoryToCopy_File_Files[i].isFile()){
                    copyFile(
                            a_DirectoryToCopy_File_Files[i],
                            copiedDirectory_File);
                } else {
                    copyDirectory(
                            a_DirectoryToCopy_File_Files[i],
                            copiedDirectory_File);
                }
            }
        } catch (IOException aIOException) {
            aIOException.printStackTrace();
            System.exit(Generic_ErrorAndExceptionHandler.IOException);
        }
    }

    public static void copy(
            File a_FileOrDirectoryToCopy_File,
            File a_DirectoryToCopyInto_File) {
        try {
            if (!a_DirectoryToCopyInto_File.mkdir()) {
                a_DirectoryToCopyInto_File.mkdirs();
            }
            if (!a_DirectoryToCopyInto_File.isDirectory()) {
                throw new IOException(
                        "Expecting File " + a_DirectoryToCopyInto_File
                        + "To be a directory in "
                        + Generic_StaticIO.class.getName() + ".copy(File,File)");
            }
            if (a_FileOrDirectoryToCopy_File.isFile()) {
//                File outputFile = new File(
//                        a_DirectoryToCopyInto_File,
//                        a_FileOrDirectoryToCopy_File.getName());
//                copyFile(
//                        a_FileOrDirectoryToCopy_File,
//                        outputFile);
                copyFile(
                        a_FileOrDirectoryToCopy_File,
                        a_DirectoryToCopyInto_File);
            } else {
                copyDirectory(
                                                a_FileOrDirectoryToCopy_File,
                        a_DirectoryToCopyInto_File);
//)
//                File[] a_DirectoryToCopy_File_Files =
//                        a_FileOrDirectoryToCopy_File.listFiles();
//                File a_DirectoryToCopyInto_1_File = new File(
//                        a_DirectoryToCopyInto_File,
//                        a_FileOrDirectoryToCopy_File.getName());
//                for (int i = 0; i < a_DirectoryToCopy_File_Files.length; i++) {
////                    copy(
////                        a_DirectoryToCopy_File_Files[i],
////                        a_DirectoryToCopyInto_1_File);
////                    }
//////                    File a_DirectoryToCopyInto_2_File = new File(
//////                            a_DirectoryToCopyInto_1_File,
//////                            a_DirectoryToCopy_File_Files[i].getName());
//////                    if (a_DirectoryToCopy_File_Files[i].isFile()) {
//////                        copyFile(
//////                                a_DirectoryToCopy_File_Files[i],
//////                                a_DirectoryToCopyInto_2_File);
//////                    } else {
//////                        if (!a_DirectoryToCopyInto_2_File.mkdir()) {
//////                            a_DirectoryToCopyInto_2_File.mkdirs();
//////                        }
//////                        copy(
//////                                a_DirectoryToCopy_File_Files[i],
//////                                a_DirectoryToCopyInto_2_File);
//////                    }
////                    File a_DirectoryToCopyInto_2_File = new File(
////                            a_DirectoryToCopyInto_1_File,
////                            a_DirectoryToCopy_File_Files[i].getName());
//                    if (a_DirectoryToCopy_File_Files[i].isFile()) {
//                        copyFile(
//                                a_DirectoryToCopy_File_Files[i],
//                                a_DirectoryToCopyInto_1_File);
//                    } else {
//                        if (!a_DirectoryToCopyInto_1_File.mkdir()) {
//                            a_DirectoryToCopyInto_1_File.mkdirs();
//                        }
//                        copy(
//                                a_DirectoryToCopy_File_Files[i],
//                                a_DirectoryToCopyInto_1_File);
//                    }
//                }
            }
        } catch (IOException aIOException) {
            aIOException.printStackTrace();
            System.exit(Generic_ErrorAndExceptionHandler.IOException);
        }
    }

    /**
     * This attempt to delete _Directory recursively.
     * @param _Directory
     * @return long[] result where result[0] is the number of directories
     * deleted; result[1] is the number of files deleted.
     */
    public static long[] deleteDirectory(File _Directory) {
        long[] result = new long[2];
        result[0] = 0L;
        result[1] = 0L;
        //long[] result_level0;
        long[] result_level1;
        File[] level0 = _Directory.listFiles();
        if (level0 != null) {
            for (int i = 0; i < level0.length; i++) {
                File[] level1 = level0[i].listFiles();
                if (level1 != null) {
                    for (int j = 0; j < level1.length; j++) {
                        if (level1[j].isDirectory()) {
                            result_level1 = deleteDirectory(level1[j]);
                            result[0] += result_level1[0];
                            result[1] += result_level1[1];
                        } else {
//                            if (!level1[j].isFile()) {
//                                int debug = 1;
//                            }
                            if (level1[j].delete()) {
                                result[1]++;
                            }
                        }
                    }
                }
                if (level0[i].isDirectory()) {
                    if (level0[i].delete()) {
                        result[0]++;
                    }
                } else {
                    //level0[i].isFile()
                    if (level0[i].delete()) {
                        result[1]++;
                    }
                }
            }
            if (level0.length == 0) {
                if (_Directory.delete()) {
                    result[0]++;
                }
            }
        }
        return result;
    }

    /**
     * Returns a newly created temporary _InputFile.
     */
    public static File createTempFile() {
        return createTempFile(null);
    }

    /**
     * Returns a newly created temporary _InputFile.
     *
     * @param parentDirectory
     *            . Default extension to nothing.
     */
    public static File createTempFile(File parentDirectory) {
        return createTempFile(parentDirectory, "", "");
    }

    /**
     * Returns a newly created temporary _InputFile.
     *
     * @param parentDirectory
     *            .
     * @param prefix
     *            If not 3 characters long, this will be padded with "x"
     *            characters.
     * @param suffix
     *            If null the _InputFile is appended with ".tmp". Default extension to
     *            nothing.
     */
    public static File createTempFile(
            File parentDirectory,
            String prefix,
            String suffix) {
        File file = null;
        while (prefix.length() < 3) {
            prefix = prefix + "x";
        }
        boolean abstractFileCreated = false;
        do {
            try {
                file = File.createTempFile(prefix + Long.toString(System.currentTimeMillis()), suffix,
                        parentDirectory);
                abstractFileCreated = true;
            } catch (IOException aIOException) {
                aIOException.printStackTrace();
                System.exit(Generic_ErrorAndExceptionHandler.IOException);
            }
        } while (!abstractFileCreated);
        file.deleteOnExit();
        return createNewFile(parentDirectory, file.getName());
    }

    /**
     * @return a File created by:
     * <code>
     * return createNewFile(new File(System.getProperty("user.dir")));
     * </code>
     */
    public static File createNewFile() {
        // return createNewFile(
        // new File( System.getProperty( "java.io.tmpdir" ) ) );
        // return createNewFile(
        // new File( System.getProperty( "user.home" ) ) );
        return createNewFile(
                new File(System.getProperty("user.dir")));
    }

    /**
     * @return a File created by:
     * <code>
     * return createNewFile(
     *         parentDirectory,
     *         new String(""),
     *         new String(""));
     * </code>
     * @param parentDirectory
     *            Default extension prefix and suffix nothing.
     */
    public static File createNewFile(File parentDirectory) {
        return createNewFile(
                parentDirectory,
                new String(""),
                new String(""));
    }

    /**
     * @return a File.
     * @param parentDirectory
     * @param prefix
     * @param suffix
     */
    public static File createNewFile(
            File parentDirectory,
            String prefix,
            String suffix) {
        File file;
        do {
            file = new File(
                    parentDirectory,
                    prefix + Long.toString(System.currentTimeMillis()) + suffix);
        } while (file.exists());
        try {
            if (new String(prefix + suffix).equalsIgnoreCase("")) {
                file.mkdir();
            } else {
                file.createNewFile();
            }
        } catch (IOException aIOException) {
            aIOException.printStackTrace();
            System.exit(Generic_ErrorAndExceptionHandler.IOException);
        }
        return file;
    }

    /**
     * Returns a newly created File which is a _InputFile if the filename. or a
     * directory.
     *
     * @param parentDirectory
     * @param filename
     */
    public static File createNewFile(
            File parentDirectory,
            String filename) {
        File file = new File(parentDirectory, filename);
        try {
            if (filename.length() > 4) {
                if (filename.charAt(filename.length() - 4) != '.') {
                    file.mkdir();
                } else {
                    file.createNewFile();
                }
            } else {
                file.mkdir();
            }
        } catch (IOException aIOException) {
            aIOException.printStackTrace();
            System.exit(Generic_ErrorAndExceptionHandler.IOException);
        }
        return file;
    }

    /**
     * Sets the syntax of aStreamTokenizer as follows:
     * <ul>
     * <li>aStreamTokenizer.resetSyntax();</li>
     * <li>aStreamTokenizer.wordChars( ',', ',' );</li>
     * <li>aStreamTokenizer.wordChars( '"', '"' );</li>
     * <li>aStreamTokenizer.wordChars( '0', '0' );</li>
     * <li>aStreamTokenizer.wordChars( '1', '1' );</li>
     * <li>aStreamTokenizer.wordChars( '2', '2' );</li>
     * <li>aStreamTokenizer.wordChars( '3', '3' );</li>
     * <li>aStreamTokenizer.wordChars( '4', '4' );</li>
     * <li>aStreamTokenizer.wordChars( '5', '5' );</li>
     * <li>aStreamTokenizer.wordChars( '6', '6' );</li>
     * <li>aStreamTokenizer.wordChars( '7', '7' );</li>
     * <li>aStreamTokenizer.wordChars( '8', '8' );</li>
     * <li>aStreamTokenizer.wordChars( '9', '9' );</li>
     * <li>aStreamTokenizer.wordChars( '.', '.' );</li>
     * <li>aStreamTokenizer.wordChars( '-', '-' );</li>
     * <li>aStreamTokenizer.wordChars( '+', '+' );</li>
     * <li>aStreamTokenizer.wordChars( 'a', 'z' );</li>
     * <li>aStreamTokenizer.wordChars( 'A', 'Z' );</li>
     * <li>aStreamTokenizer.wordChars( '\t', '\t' );</li>
     * <li>aStreamTokenizer.wordChars( ' ', ' ' );</li>
     * <li>aStreamTokenizer.eolIsSignificant( true );</li>
     * </ul>
     *
     * @param aStreamTokenizer
     *            <code>StreamTokenizer</code> thats syntax is set
     */
    public static void setStreamTokenizerSyntax1(
            StreamTokenizer aStreamTokenizer) {
        aStreamTokenizer.resetSyntax();
        // aStreamTokenizer.parseNumbers();
        aStreamTokenizer.wordChars(',', ',');
        aStreamTokenizer.wordChars('"', '"');
        // aStreamTokenizer.whitespaceChars( '"', '"' );
        // aStreamTokenizer.wordChars( '0', '9' );
        aStreamTokenizer.wordChars('0', '0');
        aStreamTokenizer.wordChars('1', '1');
        aStreamTokenizer.wordChars('2', '2');
        aStreamTokenizer.wordChars('3', '3');
        aStreamTokenizer.wordChars('4', '4');
        aStreamTokenizer.wordChars('5', '5');
        aStreamTokenizer.wordChars('6', '6');
        aStreamTokenizer.wordChars('7', '7');
        aStreamTokenizer.wordChars('8', '8');
        aStreamTokenizer.wordChars('9', '9');
        aStreamTokenizer.wordChars('.', '.');
        aStreamTokenizer.wordChars('-', '-');
        aStreamTokenizer.wordChars('+', '+');
        aStreamTokenizer.wordChars('a', 'z');
        aStreamTokenizer.wordChars('A', 'Z');
        // char[] tab = new char[1];
        // tab[0] = '\t';
        aStreamTokenizer.wordChars('\t', '\t');
        aStreamTokenizer.wordChars(' ', ' ');
        // aStreamTokenizer.ordinaryChar( ' ' );
        aStreamTokenizer.eolIsSignificant(true);
    }

    /**
     * Sets the syntax of aStreamTokenizer as follows:
     * <ul>
     * <li>aStreamTokenizer.resetSyntax();</li>
     * <li>aStreamTokenizer.wordChars( '"', '"' );</li>
     * <li>aStreamTokenizer.wordChars( '0', '0' );</li>
     * <li>aStreamTokenizer.wordChars( '1', '1' );</li>
     * <li>aStreamTokenizer.wordChars( '2', '2' );</li>
     * <li>aStreamTokenizer.wordChars( '3', '3' );</li>
     * <li>aStreamTokenizer.wordChars( '4', '4' );</li>
     * <li>aStreamTokenizer.wordChars( '5', '5' );</li>
     * <li>aStreamTokenizer.wordChars( '6', '6' );</li>
     * <li>aStreamTokenizer.wordChars( '7', '7' );</li>
     * <li>aStreamTokenizer.wordChars( '8', '8' );</li>
     * <li>aStreamTokenizer.wordChars( '9', '9' );</li>
     * <li>aStreamTokenizer.wordChars( '.', '.' );</li>
     * <li>aStreamTokenizer.wordChars( '-', '-' );</li>
     * <li>aStreamTokenizer.wordChars( '+', '+' );</li>
     * <li>aStreamTokenizer.wordChars( 'e', 'e' );</li>
     * <li>aStreamTokenizer.wordChars( 'E', 'E' );</li>
     * <li>aStreamTokenizer.wordChars( '\t', '\t' );</li>
     * <li>aStreamTokenizer.wordChars( ' ', ' ' );</li>
     * <li>aStreamTokenizer.eolIsSignificant( true );</li>
     * </ul>
     *
     * @param aStreamTokenizer
     *            <code>StreamTokenizer</code> thats syntax is set
     */
    public static void setStreamTokenizerSyntax2(
            StreamTokenizer aStreamTokenizer) {
        aStreamTokenizer.resetSyntax();
        aStreamTokenizer.wordChars('"', '"');
        aStreamTokenizer.wordChars('0', '0');
        aStreamTokenizer.wordChars('1', '1');
        aStreamTokenizer.wordChars('2', '2');
        aStreamTokenizer.wordChars('3', '3');
        aStreamTokenizer.wordChars('4', '4');
        aStreamTokenizer.wordChars('5', '5');
        aStreamTokenizer.wordChars('6', '6');
        aStreamTokenizer.wordChars('7', '7');
        aStreamTokenizer.wordChars('8', '8');
        aStreamTokenizer.wordChars('9', '9');
        aStreamTokenizer.wordChars('.', '.');
        aStreamTokenizer.wordChars('-', '-');
        aStreamTokenizer.wordChars('+', '+');
        aStreamTokenizer.wordChars('e', 'e');
        aStreamTokenizer.wordChars('E', 'E');
        aStreamTokenizer.wordChars('\t', '\t');
        aStreamTokenizer.wordChars(' ', ' ');
        aStreamTokenizer.eolIsSignificant(true);
    }

    /**
     * Sets the syntax of aStreamTokenizer as follows:
     * <ul>
     * <li>aStreamTokenizer.resetSyntax();</li>
     * <li>aStreamTokenizer.wordChars( ',', ',' );</li>
     * <li>aStreamTokenizer.wordChars( '"', '"' );</li>
     * <li>aStreamTokenizer.wordChars( '0', '0' );</li>
     * <li>aStreamTokenizer.wordChars( '1', '1' );</li>
     * <li>aStreamTokenizer.wordChars( '2', '2' );</li>
     * <li>aStreamTokenizer.wordChars( '3', '3' );</li>
     * <li>aStreamTokenizer.wordChars( '4', '4' );</li>
     * <li>aStreamTokenizer.wordChars( '5', '5' );</li>
     * <li>aStreamTokenizer.wordChars( '6', '6' );</li>
     * <li>aStreamTokenizer.wordChars( '7', '7' );</li>
     * <li>aStreamTokenizer.wordChars( '8', '8' );</li>
     * <li>aStreamTokenizer.wordChars( '9', '9' );</li>
     * <li>aStreamTokenizer.wordChars( '.', '.' );</li>
     * <li>aStreamTokenizer.wordChars( '-', '-' );</li>
     * <li>aStreamTokenizer.wordChars( '+', '+' );</li>
     * <li>aStreamTokenizer.wordChars( 'a', 'z' );</li>
     * <li>aStreamTokenizer.wordChars( 'A', 'Z' );</li>
     * <li>aStreamTokenizer.wordChars( '\t', '\t' );</li>
     * <li>aStreamTokenizer.wordChars( ' ', ' ' );</li>
     * <li>aStreamTokenizer.wordChars( '_', '_' );</li>
     * <li>aStreamTokenizer.eolIsSignificant( true );</li>
     * </ul>
     *
     * @param aStreamTokenizer
     *            <code>StreamTokenizer</code> thats syntax is set
     */
    public static void setStreamTokenizerSyntax3(
            StreamTokenizer aStreamTokenizer) {
        aStreamTokenizer.resetSyntax();
        // aStreamTokenizer.parseNumbers();
        aStreamTokenizer.wordChars(',', ',');
        aStreamTokenizer.wordChars('"', '"');
        // aStreamTokenizer.whitespaceChars( '"', '"' );
        // aStreamTokenizer.wordChars( '0', '9' );
        aStreamTokenizer.wordChars('0', '0');
        aStreamTokenizer.wordChars('1', '1');
        aStreamTokenizer.wordChars('2', '2');
        aStreamTokenizer.wordChars('3', '3');
        aStreamTokenizer.wordChars('4', '4');
        aStreamTokenizer.wordChars('5', '5');
        aStreamTokenizer.wordChars('6', '6');
        aStreamTokenizer.wordChars('7', '7');
        aStreamTokenizer.wordChars('8', '8');
        aStreamTokenizer.wordChars('9', '9');
        aStreamTokenizer.wordChars('.', '.');
        aStreamTokenizer.wordChars('-', '-');
        aStreamTokenizer.wordChars('+', '+');
        aStreamTokenizer.wordChars('a', 'z');
        aStreamTokenizer.wordChars('A', 'Z');
        // char[] tab = new char[1];
        // tab[0] = '\t';
        aStreamTokenizer.wordChars('\t', '\t');
        aStreamTokenizer.wordChars(' ', ' ');
        aStreamTokenizer.wordChars('_', '_');
        // aStreamTokenizer.ordinaryChar( ' ' );
        aStreamTokenizer.eolIsSignificant(true);
    }

    /**
     * Sets the syntax of aStreamTokenizer as follows:
     * <ul>
     * <li>aStreamTokenizer.resetSyntax();</li>
     * <li>aStreamTokenizer.wordChars( ',', ',' );</li>
     * <li>aStreamTokenizer.wordChars( '"', '"' );</li>
     * <li>aStreamTokenizer.wordChars( '0', '0' );</li>
     * <li>aStreamTokenizer.wordChars( '1', '1' );</li>
     * <li>aStreamTokenizer.wordChars( '2', '2' );</li>
     * <li>aStreamTokenizer.wordChars( '3', '3' );</li>
     * <li>aStreamTokenizer.wordChars( '4', '4' );</li>
     * <li>aStreamTokenizer.wordChars( '5', '5' );</li>
     * <li>aStreamTokenizer.wordChars( '6', '6' );</li>
     * <li>aStreamTokenizer.wordChars( '7', '7' );</li>
     * <li>aStreamTokenizer.wordChars( '8', '8' );</li>
     * <li>aStreamTokenizer.wordChars( '9', '9' );</li>
     * <li>aStreamTokenizer.wordChars( '.', '.' );</li>
     * <li>aStreamTokenizer.wordChars( '-', '-' );</li>
     * <li>aStreamTokenizer.wordChars( '+', '+' );</li>
     * <li>aStreamTokenizer.wordChars( 'a', 'z' );</li>
     * <li>aStreamTokenizer.wordChars( 'A', 'Z' );</li>
     * <li>aStreamTokenizer.eolIsSignificant( true );</li>
     * </ul>
     *
     * @param aStreamTokenizer
     *            <code>StreamTokenizer</code> thats syntax is set
     */
    public static void setStreamTokenizerSyntax4(
            StreamTokenizer aStreamTokenizer) {
        aStreamTokenizer.resetSyntax();
        aStreamTokenizer.wordChars(',', ',');
        aStreamTokenizer.wordChars('"', '"');
        aStreamTokenizer.wordChars('0', '0');
        aStreamTokenizer.wordChars('1', '1');
        aStreamTokenizer.wordChars('2', '2');
        aStreamTokenizer.wordChars('3', '3');
        aStreamTokenizer.wordChars('4', '4');
        aStreamTokenizer.wordChars('5', '5');
        aStreamTokenizer.wordChars('6', '6');
        aStreamTokenizer.wordChars('7', '7');
        aStreamTokenizer.wordChars('8', '8');
        aStreamTokenizer.wordChars('9', '9');
        aStreamTokenizer.wordChars('.', '.');
        aStreamTokenizer.wordChars('-', '-');
        aStreamTokenizer.wordChars('+', '+');
        aStreamTokenizer.wordChars('a', 'z');
        aStreamTokenizer.wordChars('A', 'Z');
        aStreamTokenizer.eolIsSignificant(true);
    }

    /**
     * Sets the syntax of aStreamTokenizer as follows:
     * <ul>
     * <li>aStreamTokenizer.resetSyntax();</li>
     * <li>aStreamTokenizer.wordChars( ',', ',' );</li>
     * <li>aStreamTokenizer.wordChars( '"', '"' );</li>
     * <li>aStreamTokenizer.wordChars( '0', '0' );</li>
     * <li>aStreamTokenizer.wordChars( '1', '1' );</li>
     * <li>aStreamTokenizer.wordChars( '2', '2' );</li>
     * <li>aStreamTokenizer.wordChars( '3', '3' );</li>
     * <li>aStreamTokenizer.wordChars( '4', '4' );</li>
     * <li>aStreamTokenizer.wordChars( '5', '5' );</li>
     * <li>aStreamTokenizer.wordChars( '6', '6' );</li>
     * <li>aStreamTokenizer.wordChars( '7', '7' );</li>
     * <li>aStreamTokenizer.wordChars( '8', '8' );</li>
     * <li>aStreamTokenizer.wordChars( '9', '9' );</li>
     * <li>aStreamTokenizer.wordChars( '.', '.' );</li>
     * <li>aStreamTokenizer.wordChars( '-', '-' );</li>
     * <li>aStreamTokenizer.wordChars( '+', '+' );</li>
     * <li>aStreamTokenizer.wordChars( 'a', 'z' );</li>
     * <li>aStreamTokenizer.wordChars( 'A', 'Z' );</li>
     * <li>aStreamTokenizer.wordChars( '\t', '\t' );</li>
     * <li>aStreamTokenizer.wordChars( ' ', ' ' );</li>
     * <li>aStreamTokenizer.wordChars( ':', ':' );</li>
     * <li>aStreamTokenizer.wordChars( '/', '/' );</li>
     * <li>aStreamTokenizer.wordChars( '_', '_' );</li>
     * </ul>
     *
     * @param aStreamTokenizer
     *            <code>StreamTokenizer</code> thats syntax is set
     */
    public static void setStreamTokenizerSyntax5(
            StreamTokenizer aStreamTokenizer) {
        aStreamTokenizer.resetSyntax();
        // aStreamTokenizer.parseNumbers();
        aStreamTokenizer.wordChars(',', ',');
        aStreamTokenizer.wordChars('"', '"');
        // aStreamTokenizer.whitespaceChars( '"', '"' );
        // aStreamTokenizer.wordChars( '0', '9' );
        aStreamTokenizer.wordChars('0', '0');
        aStreamTokenizer.wordChars('1', '1');
        aStreamTokenizer.wordChars('2', '2');
        aStreamTokenizer.wordChars('3', '3');
        aStreamTokenizer.wordChars('4', '4');
        aStreamTokenizer.wordChars('5', '5');
        aStreamTokenizer.wordChars('6', '6');
        aStreamTokenizer.wordChars('7', '7');
        aStreamTokenizer.wordChars('8', '8');
        aStreamTokenizer.wordChars('9', '9');
        aStreamTokenizer.wordChars('.', '.');
        aStreamTokenizer.wordChars('-', '-');
        aStreamTokenizer.wordChars('+', '+');
        aStreamTokenizer.wordChars('a', 'z');
        aStreamTokenizer.wordChars('A', 'Z');
        // char[] tab = new char[1];
        // tab[0] = '\t';
        aStreamTokenizer.wordChars('\t', '\t');
        aStreamTokenizer.wordChars(' ', ' ');
        aStreamTokenizer.wordChars(':', ':');
        aStreamTokenizer.wordChars('/', '/');
        aStreamTokenizer.wordChars('_', '_');
        // aStreamTokenizer.ordinaryChar( ' ' );
        aStreamTokenizer.eolIsSignificant(true);
    }

    /**
     * @return File directory in which object with a_ID is (to be) stored within
     * directory0.
     * @param directory
     * @param a_ID The ID of the object to be stored
     * @param max_ID The maximum number of objects to be stored
     * @param range The number of objects stored per directory
     */
    public static File getObjectDirectory(
            File directory,
            long a_ID,
            long max_ID,
            long range) {
        long diff = range;
        while (max_ID / (double) diff > 1) {
            //while (max_ID / diff > 1) {
            diff *= range;
        }
        long min = 0;
        Object[] min_directory = getObjectDirectory(
                a_ID,
                min,
                diff,
                directory);
        diff /= range;
        while (diff > 1L) {
            min_directory = getObjectDirectory(
                    a_ID,
                    (Long) min_directory[1],
                    diff,
                    (File) min_directory[0]);
            diff /= range;
        }
        return (File) min_directory[0];
    }

    private static Object[] getObjectDirectory(
            long a_ID,
            long min,
            long diff,
            File parent_File) {
        Object[] result = new Object[2];
        long mint = min;
        long maxt = min + diff - 1;
        String directory_String;
        File directory = null;
        boolean found = false;
        while (!found) {
            if (a_ID >= mint && a_ID <= maxt) {
                found = true;
                directory_String = new String("" + mint + "_" + maxt);
                directory = new File(
                        parent_File,
                        directory_String);
            } else {
                mint += diff;
                maxt += diff;
            }
        }
        result[0] = directory;
        result[1] = mint;
        return result;
    }

    public static File initialiseArchive(
            File directory,
            long range) {
        File result;
        long start = 0;
        long end = range - 1;
        result = new File(
                directory,
                "" + start + "_" + end + "/" + start);
        result.mkdirs();
        return result;
    }

    public static File initialiseArchiveNoLeaf(
            File directory,
            long range) {
        File result;
        long start = 0;
        long end = range - 1;
        result = new File(
                directory,
                "" + start + "_" + end);
        result.mkdirs();
        return result;
    }

    public static Long getHighestLeaf(
            File directory,
            String underscore) {
        Long result;
        File[] archiveFiles = directory.listFiles();
        if (archiveFiles == null) {
            return -1L;
        }
        if (archiveFiles.length == 0) {
            return -1L;
        } else {
            TreeMap<Long, File> numericalOrderFiles_TreeMap =
                    getNumericalOrderFilesWithNumericalUnderscoreFilenames_TreeMap(
                    archiveFiles,
                    underscore);
            File last_File = numericalOrderFiles_TreeMap.lastEntry().getValue();
            if (last_File.getName().contains(underscore)) {
                return getHighestLeaf(
                        last_File,
                        underscore);
            } else {
                result = new Long(last_File.getName());
            }
        }
        return result;
    }

    public static File growArchiveBase(
            File directory,
            long range) {
        String underscore = "_";
        File[] archiveFiles = directory.listFiles();
        TreeMap<Long, File> numericalOrderFiles_TreeMap = getNumericalOrderFilesWithNumericalUnderscoreFilenames_TreeMap(
                archiveFiles,
                underscore);
        File first_File = numericalOrderFiles_TreeMap.firstEntry().getValue();
        String directoryName0 = getFilename(
                directory,
                first_File);
        String[] split0 = directoryName0.split(underscore);
        //long start0 = new Long(split0[0]).longValue();
        long end0 = new Long(split0[1]).longValue();
        long newRange = range * (end0 + 1L);
        // Create new top directory and move in existing files
        File newTopOfArchive0 = new File(
                directory,
                "" + 0 + "_" + (newRange - 1L));
        File newTopOfArchive = new File(newTopOfArchive0.getPath());
        newTopOfArchive0.mkdir();
        for (int i = 0; i < archiveFiles.length; i++) {
            File newPath = new File(
                    newTopOfArchive,
                    archiveFiles[i].getName());
            archiveFiles[i].renameTo(newPath);
        }
        // Create new lower directories for next ID;
        Long next_ID = getHighestLeaf(
                directory,
                underscore);
        next_ID++;
        File newHighestLeaf_Directory = new File(
                getObjectDirectory(
                newTopOfArchive, next_ID, next_ID, range),
                "" + next_ID);
        newHighestLeaf_Directory.mkdirs();
        return newTopOfArchive0;
    }

    public static File growArchiveBase(
            File directory,
            long range,
            long next_ID) {
        String underscore = "_";
        File[] archiveFiles = directory.listFiles();
        TreeMap<Long, File> numericalOrderFiles_TreeMap = getNumericalOrderFilesWithNumericalUnderscoreFilenames_TreeMap(
                archiveFiles,
                underscore);
        File first_File = numericalOrderFiles_TreeMap.firstEntry().getValue();
        String directoryName0 = getFilename(
                directory,
                first_File);
        String[] split0 = directoryName0.split(underscore);
        //long start0 = new Long(split0[0]).longValue();
        long end0 = new Long(split0[1]).longValue();
        long newRange = range * (end0 + 1L);
        // Create new top directory and move in existing files
        File newTopOfArchive0 = new File(
                directory,
                "" + 0 + "_" + (newRange - 1L));
        File newTopOfArchive = new File(newTopOfArchive0.getPath());
        newTopOfArchive0.mkdir();
        for (int i = 0; i < archiveFiles.length; i++) {
            File newPath = new File(
                    newTopOfArchive,
                    archiveFiles[i].getName());
            archiveFiles[i].renameTo(newPath);
        }
        File newHighestLeaf_Directory = new File(
                getObjectDirectory(
                newTopOfArchive, next_ID, next_ID, range),
                "" + next_ID);
        newHighestLeaf_Directory.mkdirs();
        return newTopOfArchive0;
    }

    public static File addToArchive(
            File directory0,
            long range) {
        String underscore = "_";
        Long next_ID = getHighestLeaf(
                directory0,
                underscore);
        next_ID++;
        File newHighestLeaf_Directory = new File(
                getObjectDirectory(
                directory0,
                next_ID,
                //next_ID + range - 2,
                next_ID + 1,
                range),
                "" + next_ID);
        String filename = getFilename(
                directory0,
                newHighestLeaf_Directory);
        File potentialNewTopDirectory = new File(
                directory0,
                filename);
        // Test range
        String[] splitnew = filename.split(underscore);
        long startnew = new Long(splitnew[0]).longValue();
        long endnew = new Long(splitnew[1]).longValue();
        long rangenew = endnew - startnew;
        String[] archiveFiles = directory0.list();
        String[] splitold = archiveFiles[0].split(underscore);
        long startold = new Long(splitold[0]).longValue();
        long endold = new Long(splitold[1]).longValue();
        long rangeold = endold - startold;
        if (rangenew > rangeold) {
            growArchiveBase(directory0, range);
        }
        newHighestLeaf_Directory.mkdirs();
        return newHighestLeaf_Directory;
    }

    public static File addToArchive(
            File directory0,
            long range,
            long next_ID) {
        String underscore = "_";
        File newHighestLeaf_Directory = new File(
                getObjectDirectory(
                directory0,
                next_ID,
                //next_ID + range - 2,
                next_ID + 1,
                range),
                "" + next_ID);
        String filename = getFilename(
                directory0,
                newHighestLeaf_Directory);
        File potentialNewTopDirectory = new File(
                directory0,
                filename);
        // Test range
        String[] splitnew = filename.split(underscore);
        long startnew = new Long(splitnew[0]).longValue();
        long endnew = new Long(splitnew[1]).longValue();
        long rangenew = endnew - startnew;
        String[] archiveFiles = directory0.list();
        String[] splitold = archiveFiles[0].split(underscore);
        long startold = new Long(splitold[0]).longValue();
        long endold = new Long(splitold[1]).longValue();
        long rangeold = endold - startold;
        if (rangenew > rangeold) {
            growArchiveBase(directory0, range, next_ID);
        }
        newHighestLeaf_Directory.mkdirs();
        return newHighestLeaf_Directory;
    }
//
//        String[] directory0list = directory0.list();
//        boolean topDirectoryExists;
//        for (int i = 0; i < directory0list.length; i ++){
//            if (directory0list.equals(filename)){
//
//            }
//        TreeMap<Long, File> numericalOrderFiles_TreeMap = getNumericalOrderFilesWithNumericalUnderscoreFilenames_TreeMap(
//                archiveFiles,
//                underscore);
//
//
//
//
//        newHighestLeaf_Directory.mkdirs();
//
//
//
//        File last_File = numericalOrderFiles_TreeMap.lastEntry().getValue();
//
//        String filename2
//
//
//
//        boolean result = false;
//        String underscore = new String("_");
//        File[] list = directory.listFiles();
//        if (list != null) {
//            if (list.length > 0) {
//                if (list.length == range) {
//                    if (directory0 == directory) {
//                        growArchive(
//                                directory0,
//                                directory,
//                                range);
//                        return true;
//                    }
//                    boolean attemptResult = addToArchive(
//                            directory0,
//                            directory.getParentFile(),
//                            range);
////                    if (attemptResult) {
////                        return true;
////                    }
//                    return attemptResult;
//                }
//                TreeMap<Long, File> numericalOrderFiles_TreeMap = getNumericalOrderFilesWithNumericalUnderscoreFilenames_TreeMap(
//                        list,
//                        underscore);
//                File lastFile = numericalOrderFiles_TreeMap.lastEntry().getValue();
//                String filename = getFilename(
//                        directory,
//                        lastFile);
//                if (filename.contains(underscore)) {
//                    boolean attemptResult = addToArchive(
//                            directory0,
//                            //lastFile,
//                            lastFile.getParentFile(),
//                            range);
//                    if (attemptResult) {
//                        return true;
//                    }
//                    if (directory0 == directory) {
//                        growArchive(
//                                directory0,
//                                directory,
//                                range);
//                        return true;
//                    }
//                    File[] listParentFiles = lastFile.getParentFile().listFiles();
//                    TreeMap<Long, File> numericalOrderParentFiles_TreeMap = getNumericalOrderFilesWithNumericalUnderscoreFilenames_TreeMap(
//                            listParentFiles,
//                            underscore);
//                    Entry<Long, File> lastParentFile_Entry = numericalOrderParentFiles_TreeMap.lastEntry();
//                    Entry<Long, File> penultimateParentFile_Entry = numericalOrderParentFiles_TreeMap.higherEntry(lastParentFile_Entry.getKey());
//                    File penultimateFile = penultimateParentFile_Entry.getValue();
//                    String penultimateFilename = penultimateFile.getName();
//                    String[] split = penultimateFilename.split(underscore);
//                    long start0 = new Long(split[0]).longValue();
//                    long end0 = new Long(split[1]).longValue();
//                    long currentRange = end0 - start0 + 1;
//                    File newTopDirectory = new File(
//                            directory,
//                            "" + (start0 + currentRange) + "_"
//                            + (end0 + currentRange));
//                    newTopDirectory.mkdir();
//                    File lastParentFile = lastParentFile_Entry.getValue();
//                    File newPath = new File(
//                            newTopDirectory,
//                            lastParentFile.getName());
//                    lastParentFile.renameTo(newPath);
//                    //}
////                    while (currentRange > range) {
////                        currentRange /= range;
////                        newTopDirectory = new File(
////                                newTopDirectory,
////                                "" + start0 + "_" + (start0 + currentRange));
////                    }
////                    newTopDirectory = new File(
////                            newTopDirectory,
////                            "" + (start0 + currentRange));
////                    newTopDirectory.mkdirs();
//                    return true;
//                } else {
//                    if (list.length < range) {
//                        long fileID = new Long(filename);
//                        fileID++;
//                        File newTopDirectory = new File(
//                                directory,
//                                "" + fileID);
//                        newTopDirectory.mkdirs();
//                        return true;
//                    } else {
//                        growArchive(
//                                directory0,
//                                directory0,
//                                range);
////                    return addToArchive(
////                            directory.getParentFile(),
////                            range);
//                    }
//                }
//            } else {
//                if (directory == directory0) {
//                    int a = 0;
//                    a++;
//                }
//                String filename = getFilename(
//                        directory.getParentFile(),
//                        directory);
//                if (filename.contains(underscore)) {
//                    boolean attemptResult = addToArchive(
//                            directory0,
//                            directory.getParentFile(),
//                            range);
//                    if (attemptResult) {
//                        return result;
//                    }
//                    String[] split = filename.split(underscore);
//                    long start0 = new Long(split[0]).longValue();
//                    long end0 = new Long(split[1]).longValue();
//                    long currentRange = end0 - start0 + 1;
//                    File newTopDirectory = new File(
//                            directory,
//                            "" + (start0 + currentRange) + "_"
//                            + (end0 + currentRange));
//                    while (currentRange > range) {
//                        currentRange /= range;
//                        newTopDirectory = new File(
//                                newTopDirectory,
//                                "" + start0 + "_" + (start0 + currentRange));
//                    }
//                    newTopDirectory = new File(
//                            newTopDirectory,
//                            "" + (start0 + currentRange));
//                    newTopDirectory.mkdirs();
//                    return true;
//                }
//            }
//        }
//        return result;
//    }

//    public static File growArchive(
//            File directory0,
//            File directory,
//            long range) {
//        String underscore = new String("_");
//        File[] archiveFiles = directory.listFiles();
//        TreeMap<Long, File> numericalOrderFiles_TreeMap = getNumericalOrderFilesWithNumericalUnderscoreFilenames_TreeMap(
//                archiveFiles,
//                underscore);
//        File first_File = numericalOrderFiles_TreeMap.firstEntry().getValue();
//        String directoryName0 = getFilename(
//                directory,
//                first_File);
//        String[] split0 = directoryName0.split(underscore);
//        long start0 = new Long(split0[0]).longValue();
//        long end0 = new Long(split0[1]).longValue();
//        //if (archiveFiles.length > 1) {
//        if (archiveFiles.length == range) {
//            if (directory != directory0) {
//                return growArchive(
//                        directory0,
//                        directory.getParentFile(),
//                        range);
//            } else {
//                File newTopOfArchive = new File(
//                        directory,
//                        "" + start0 + "_" + ((range + 1) * end0));
//                newTopOfArchive.mkdir();
//                for (int i = 0; i < archiveFiles.length; i++) {
//                    File newPath = new File(
//                            newTopOfArchive,
//                            archiveFiles[i].getName());
//                    archiveFiles[i].renameTo(newPath);
//                }
//                long newRange0 = ((range + 1) * end0) - start0;
//                start0 += newRange0;
//                end0 += newRange0;
//                newTopOfArchive = new File(
//                        directory,
//                        "" + start0 + "_" + end0);
//                newTopOfArchive.mkdir();
//                long newRange = newRange0;
//                while (newRange < range) {
//                    newRange /= range;
//                    newTopOfArchive = new File(
//                            newTopOfArchive,
//                            "" + start0 + "_" + (start0 + newRange));
//                }
//                newTopOfArchive = new File(
//                        newTopOfArchive,
//                        "" + start0);
//                newTopOfArchive.mkdir();
//            }
//        } else {
//            File last_File = numericalOrderFiles_TreeMap.lastEntry().getValue();
//            String directoryNameNMinusOne = getFilename(
//                    directory,
//                    last_File);
//            String[] splitNMinusOne = directoryNameNMinusOne.split(underscore);
//            long startNMinusOne = new Long(splitNMinusOne[0]).longValue();
//            long endNMinusOne = new Long(splitNMinusOne[1]).longValue();
//            long newRangeNMinusOne = endNMinusOne - startNMinusOne + 1;
//            long startN = startNMinusOne + newRangeNMinusOne;
//            long endN = endNMinusOne + newRangeNMinusOne;
//            File newTopOfArchive = new File(
//                    directory,
//                    "" + startN + "_" + endN);
//            long newRange = newRangeNMinusOne;
//            while (newRange < range) {
//                newRange /= range;
//                newTopOfArchive = new File(
//                        newTopOfArchive,
//                        "" + startN + "_" + (startN + newRange));
//            }
//            newTopOfArchive = new File(
//                    newTopOfArchive,
//                    "" + startN);
//            newTopOfArchive.mkdirs();
//            return newTopOfArchive;
//        }
//        //}
//        return null;
//    }
//    public static File createArchive(
//            File directory,
//            long range) {
//        File[] archives = directory.listFiles();
//        File newTopOfArchiveDirectory;
//        if (archives != null) {
//            if (archives.length == range) {
//                String filename = getFilename(
//                        directory,
//                        archives[archives.length - 1]);
//                String[] split = filename.split("_");
//                long max_ID;
//                if (split.length == 1) {
//                    max_ID = new Long(split[0]);
//                } else {
//                    max_ID = new Long(filename.split("_")[1]);
//                }
//                max_ID += range;
//                long next_ID = max_ID - archives.length + 1;
//                if (directory.getParentFile().list().length < range - 1) {
//                    System.out.println("directory.getParentFile().list().length < range -1");
//                } else {
//                    System.out.println("directory.getParentFile().list().length !< range -1");
//                }
//                growArchive(directory, max_ID, range);
//                File objectParentDirectory = getObjectDirectory(
//                        directory.getParentFile(), next_ID, max_ID, range);
//                newTopOfArchiveDirectory = new File(
//                        objectParentDirectory,
//                        "" + next_ID);
//                //objectDirectory.mkdirs();
//                //newTopOfArchiveDirectory = new File(directory,
//                //        getFilename(directory, objectDirectory));
//                //directory = growArchive(directory.getParentFile(), max_ID, range);
//                //newTopOfArchiveDirectory = createArchive(directory, range);
//                // Where is new top of archive?
//                //newTopOfArchiveDirectory = getObjectDirectory(
//                //        directory, max_ID, max_ID, range);
//            } else {
//                if (archives.length == 0) {
////                    String filename = getFilename(
////                            directory,
////                            archives[archives.length - 1]);
////                    String[] split = filename.split("_");
////                    long a_ID;
////                    if (split.length == 1) {
////                        a_ID = new Long(split[0]);
////                    } else {
////                        a_ID = new Long(filename.split("_")[1]);
////                    }
//
//
//
//                    long start = archives.length * range;
//                    long end = start + range - 1;
//                    newTopOfArchiveDirectory = new File(
//                            directory, "" + start + "_" + end + "/" + start);
//                } else {
//                    String filename = getFilename(
//                            directory,
//                            archives[archives.length - 1]);
//                    String[] split = filename.split("_");
//                    long a_ID;
//                    if (split.length == 1) {
//                        a_ID = new Long(split[0]);
//                    } else {
//                        a_ID = new Long(filename.split("_")[1]);
//                    }
//                    a_ID += 1;
//                    newTopOfArchiveDirectory = new File(
//                            directory, "" + a_ID);
//                }
//            }
//        } else {
//            newTopOfArchiveDirectory = new File(directory.toString() + "/0_" + (range - 1) + "/0");
//        }
//        newTopOfArchiveDirectory.mkdirs();
//        // Attempt to slow things down as for some reason file system is getting
//        // messy...
//        System.out.println(
//                "new directory " + newTopOfArchiveDirectory
//                + " exists == " + newTopOfArchiveDirectory.exists());
////        for(int i = 0; i < 10000; i ++){
////            System.out.print(".");
////        }
//        return newTopOfArchiveDirectory;
//    }
    /**
     * @param directory
     * @param a_File
     * @return The name of the file or directory in directory in the path of
     * a_File.
     */
    public static String getFilename(
            File directory,
            File a_File) {
        String result;
        int beginIndex = directory.getAbsolutePath().length() + 1;
        String fileSeparator = System.getProperty("file.separator");
        /*
         * A feature in Java means splitting strings with "\" does not work as
         * might be expected and the regexp needs changing to "\\\\"
         */
        String regexp = new String("\\");
        //System.out.println("regexp " + regexp);
        if (fileSeparator.equals(regexp)) {
            fileSeparator = "\\\\";
        }
        //System.out.println("fileSeparator " + fileSeparator);
        //String newTopOfArchiveDirectoryName = (objectDirectoryFile.getAbsolutePath().substring(beginIndex)).split(System.getProperty("file.separator"))[0];
        //String newTopOfArchiveDirectoryName = (objectDirectoryFile.getPath().substring(beginIndex)).split("\\")[0];
        //String newTopOfArchiveDirectoryName = (objectDirectoryFile.getPath().substring(beginIndex)).split("/")[0];
        result = (a_File.getPath().substring(beginIndex)).split(fileSeparator)[0];
        //System.out.println("result " + result);
        return result;
    }

    /**
     * {@code getNumericalOrderFilesWithNumericalFilenames_TreeMap(
     *       File[] files) {
     *   TreeMap<Long, File> result = new TreeMap<Long, File>();
     *   for (int i = 0; i < files.length; i++) {
     *       result.put(
     *               new Long(files[i].getName()),
     *               files[i]);
     *   }
     *   return result;
     * }
     * }
     * @param files
     * @return TreeMap<Long, File>
     */
    public static TreeMap<Long, File> getNumericalOrderFilesWithNumericalFilenames_TreeMap(
            File[] files) {
        TreeMap<Long, File> result = new TreeMap<Long, File>();
        for (int i = 0; i < files.length; i++) {
            result.put(
                    new Long(files[i].getName()),
                    files[i]);
        }
        return result;
    }

    /**
     * If needed to order also by a number after the underscore then a new 
     * method is needed. 
     * @param files
     * @param underscore = "_"
     * @return TreeMap<Long, File>
     */
    public static TreeMap<Long, File> getNumericalOrderFilesWithNumericalUnderscoreFilenames_TreeMap(
            File[] files,
            String underscore) {
        TreeMap<Long, File> result = new TreeMap<Long, File>();
        String filename;
        String[] split;
        for (int i = 0; i < files.length; i++) {
            filename = files[i].getName();
            split = filename.split("_");
            long start = new Long(split[0]).longValue();
            result.put(
                    start,
                    files[i]);
        }
        return result;
    }
}
